# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import re
import copy
import logging
from .protobuf_helper import ProtobufHelper
from nn_meter.utils.import_package import try_import_tensorflow
logging = logging.getLogger("nn-Meter")


class FrozenPbParser:
    def __init__(self, pb_file):
        tf = try_import_tensorflow()
        f = open(pb_file, "rb")
        graph = tf.compat.v1.GraphDef()
        graph.ParseFromString(f.read())

        self.graph = graph

    @staticmethod
    def strip_useless_nodes(model_graph):
        """
        Remove nodes that does not matter with the predict or the structure of model,
        including following types:
        - weights for ops
        - attributes for ops

        Parameters
        ----------
        model_graph : ModelGraph
            the graph holder
        """
        stripped_nodes_type_all = []
        stripped_nodes_type = ["Const", "Identity"]
        stripped_nodes_keywords = [
            "/weight",
            "/weight/read",
            "/ReadVariableOp",
            "/kernel",
            "/gamma",
            "/beta",
            "/moving_mean",
            "/moving_variance",
            "/bias",
            "/reduction_indices",
            "/shape",
            "/split_dim",
            "/axis",
        ]
        graph = model_graph.get_graph()
        removed_node = []
        for key, value in graph.items():
            if "attr" in value.keys():
                if value["attr"]["type"] in stripped_nodes_type:
                    for kw in stripped_nodes_keywords:
                        if kw in key:
                            removed_node.append(key)
                            break
                if value["attr"]["type"] in stripped_nodes_type_all:
                    removed_node.append(key)

        for key in removed_node:
            del graph[key]

        model_graph.refresh()

    @staticmethod
    def fix_split_naming(model_graph):
        """
        TensorFlow is using "NODE_NAME:NUMBER"  for example "split:0", "split:1"
        as a notation to oredered outputs,
        such notation will make the edge not able to connect. We patch it by using the list
        to store the name and keep the sequence.

        Parameters
        ----------
        model_graph : ModelGraph
            the graph holder
        """
        graph = model_graph.get_graph()
        graph_nodes = copy.deepcopy(list(graph.keys()))
        remove_node_list = []
        for graph_node in graph_nodes:
            if graph_node in graph.keys():
                if "attr" in graph[graph_node].keys():
                    if (
                        graph[graph_node]["attr"]["type"] == "Split"
                        and ":" not in graph_node
                    ):
                        logging.info("Find split main node %s." % graph_node)
                        split_node_name = graph_node
                        for node_name in graph.keys():
                            idx = re.findall(r"%s:(\d+)" % split_node_name, node_name)
                            if len(idx) > 0:
                                idx = int(idx[0])
                                logging.info("Find split child node %s." % node_name)
                                graph[graph_node]["outbounds"] += graph[node_name][
                                    "outbounds"
                                ]
                                graph[graph[node_name]["outbounds"][0]]["inbounds"] += [
                                    graph_node
                                ]
                                remove_node_list.append(node_name)

        for node in remove_node_list:
            del graph[node]

        model_graph.refresh()

    def fetch_attr_to_dict(self, node):
        """
        Tensorflow store some of the attributes as a node connect to the tensor.
        We fetch the attribute from those noed to a dict.

        Parameters
        ----------
        node : Protobuf.node
            The protobuf node of the frozen pb.
        """

        attr_dict = {}

        attr_as_node = {
            "Split": {
                "node_name": lambda x: x + "/split_dim",
                "attr_name": "split_dim",
                "node_value": lambda x: ProtobufHelper.get_tensor_value(x),
            },
            "Mean": {
                "node_name": lambda x: x + "/reduction_indices",
                "attr_name": "reduction_indices",
                "node_value": lambda x: ProtobufHelper.pkg42dec(x.tensor_content),
            },
            "Reshape": {
                "node_name": lambda x: x + "/shape",
                "attr_name": "shape",
                "node_value": lambda x: ProtobufHelper.pkg42dec(x.tensor_content),
            },
            "Concat": {
                "node_name": lambda x: x + "/axis",
                "attr_name": "axis",
                "node_value": lambda x: ProtobufHelper.get_tensor_value(x),
            },
            "ConcatV2": {
                "node_name": lambda x: x + "/axis",
                "attr_name": "axis",
                "node_value": lambda x: ProtobufHelper.get_tensor_value(x),
            },
            "Const": {
                "node_name": lambda x: x,
                "attr_name": "constant",
                "node_value": lambda x: ProtobufHelper.get_tensor_value(x),
            },
            "Pack": {
                "node_name": lambda x: x + r"/(\d)",
                "regex": True,
                "attr_name": "constant",
                "node_value": lambda x: ProtobufHelper.get_tensor_value(x),
            },
        }

        list_i_nodes = ["dilations", "strides", "ksize"]
        str_nodes = ["padding", "data_format"]

        for attr_name in node.attr.keys():
            if attr_name in list_i_nodes:
                attr_dict[attr_name] = [int(a) for a in node.attr[attr_name].list.i]
                continue

            if attr_name in str_nodes:
                attr_dict[attr_name] = node.attr[attr_name].s
                continue

            if attr_name == "value":
                shape = []
                for dim in node.attr[attr_name].tensor.tensor_shape.dim:
                    shape.append(dim.size)
                attr_dict["tensor_shape"] = list(map(int, shape))
                continue

            if attr_name == "shape":
                shape = []
                for dim in node.attr[attr_name].shape.dim:
                    shape.append(dim.size)
                attr_dict["shape"] = list(map(int, shape))
                continue

        if node.op in attr_as_node.keys():
            for target_node in self.graph.node:
                if "regex" in attr_as_node[node.op].keys():
                    node_attr = re.findall(
                        attr_as_node[node.op]["node_name"](node.name), target_node.name
                    )
                    if len(node_attr) > 0:
                        logging.info("Find regex matching node %s" % node.name)
                        for attr_name in target_node.attr.keys():
                            if (
                                attr_name == "value"
                                and "weight" not in node.name
                                and "BatchNorm" not in node.name
                                and "kernel" not in node.name
                            ):
                                node_attr_name = attr_as_node[node.op]["attr_name"]
                                if node_attr_name not in attr_dict.keys():
                                    attr_dict[node_attr_name] = []
                                attr_dict[node_attr_name].append(
                                    copy.deepcopy(
                                        attr_as_node[node.op]["node_value"](
                                            target_node.attr[attr_name].tensor
                                        )
                                    )
                                )
                else:
                    if target_node.name == attr_as_node[node.op]["node_name"](
                        node.name
                    ):
                        for attr_name in target_node.attr.keys():
                            if (
                                attr_name == "value"
                                and "weight" not in node.name
                                and "BatchNorm" not in node.name
                                and "kernel" not in node.name
                            ):
                                attr_dict[
                                    attr_as_node[node.op]["attr_name"]
                                ] = copy.deepcopy(
                                    attr_as_node[node.op]["node_value"](
                                        target_node.attr[attr_name].tensor
                                    )
                                )

        return attr_dict

    def parse_graph(self, model_graph):
        """
        Parse a frozen protobuf file from tensorflow to graph IR

        Parameters
        ----------
        model_graph : ModelGraph
            The Graph IR holder.
        """

        for node in self.graph.node:
            model_graph.node(str(node.name), list(map(str, node.input)))
            model_graph.set_node_attr(
                node.name,
                {
                    "name": str(node.name),
                    "type": str(node.op),
                    "output_shape": [], # This will be filled later
                    "attr": self.fetch_attr_to_dict(node),
                },
            )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import sys
import logging
import argparse
from .registry import register_module_cli, unregister_module_cli
from .predictor import list_latency_predictors_cli, apply_latency_predictor_cli, get_nnmeter_ir_cli
from .builder import list_backends_cli, list_kernels_cli, list_operators_cli, list_special_testcases_cli, \
    test_backend_connection_cli, create_workspace_cli


def nn_meter_info(args):
    if args.list_predictors:
        list_latency_predictors_cli()
    elif args.list_backends:
        list_backends_cli()
    elif args.list_kernels:
        list_kernels_cli()
    elif args.list_operators:
        list_operators_cli()
    elif args.list_testcases:
        list_special_testcases_cli()
    else:
        logging.keyinfo('please run "nn-meter {positional argument} --help" to see nn-meter guidance')


def nn_meter_cli():
    parser = argparse.ArgumentParser('nn-meter', description='please run "nn-meter {positional argument} --help" to see nn-meter guidance')
    parser.set_defaults(func=nn_meter_info)

    # optional arguments
    parser.add_argument(
        "-v", "--verbose", 
        help="increase output verbosity",
        action="store_true"
    )
    parser.add_argument(
        '--list-predictors',
        help='list all supported predictors',
        action='store_true',
        default=False
    )
    parser.add_argument(
        '--list-backends',
        help='list all supported backends',
        action='store_true',
        default=False
    )
    parser.add_argument(
        '--list-kernels',
        help='list all supported kernels when building kernel predictors',
        action='store_true',
        default=False
    )
    parser.add_argument(
        '--list-operators',
        help='list all supported operators when building fusion rule test cases',
        action='store_true',
        default=False
    )
    parser.add_argument(
        '--list-testcases',
        help='list all supported special test cases when building fusion rule test cases',
        action='store_true',
        default=False
    )

    # create subparsers for args with sub values
    subparsers = parser.add_subparsers()

    # Usage 1: predict latency for models
    # Usage: nn-meter predict --predictor <hardware> --tensorflow <pb-file>
    lat_pred = subparsers.add_parser('predict', aliases=['lat_pred'], help='apply latency predictor for testing model')
    lat_pred.add_argument(
        "--predictor",
        type=str,
        help="name of target predictor (hardware)"
    )
    lat_pred.add_argument(
        "--predictor-version",
        type=float,
        help="the version of the latency predictor (if not specified, use the lateast version)",
        default=None
    )
    model_type = lat_pred.add_mutually_exclusive_group()
    model_type.add_argument(
        "--tensorflow",
        type=str,
        help="path to input Tensorflow model (*.pb file or floder)"
    )
    model_type.add_argument(
        "--onnx",
        type=str,
        help="path to input ONNX model (*.onnx file or floder)"
    )
    model_type.add_argument(
        "--nn-meter-ir",
        type=str,
        help="path to input nn-Meter IR model (*.json file or floder)"
    )
    model_type.add_argument(
        "--torchvision",        # --torchvision only can support the model object. The argument specifies 
        type=str,               # the name of the model, and we will look for the model in torchvision model zoo.
        nargs='+',
        help="name of the input torch model from the torchvision model zoo"
    )
    lat_pred.set_defaults(func=apply_latency_predictor_cli)

    # Usage 2: get nn-meter-ir model from tensorflow pbfile or onnx file
    # Usage: nn-meter get_ir --tensorflow <pb-file>
    get_ir = subparsers.add_parser(
        'get_ir', 
        help='specify a model type to convert to nn-meter ir graph'
    )
    model_type = get_ir.add_mutually_exclusive_group()
    model_type.add_argument(
        "--tensorflow",
        type = str,
        help="path to input Tensorflow model (*.pb)"
    )
    model_type.add_argument(
        "--onnx",
        type=str,
        help="path to input ONNX model (*.onnx)"
    )
    get_ir.add_argument(
        "-o", "--output",
        type=str,
        help="path to save the output nn-meter ir graph for tensorflow and onnx (*.json), default to be /path/to/input/file/<input_file_name>_ir.json"
    )
    get_ir.set_defaults(func=get_nnmeter_ir_cli)

    # Usage 3: create workspace folder for nn-Meter builder 
    # Usage: nn-meter create --tflite-workspace <path/to/workspace>
    create_workspace = subparsers.add_parser(
        'create', 
        help='create a workspace folder for nn-Meter builder'
    )
    platform = create_workspace.add_mutually_exclusive_group()
    platform.add_argument(
        "--tflite-workspace",
        type=str,
        help="path to place a tflite workspace for nn-Meter builder"
    )
    platform.add_argument(
        "--openvino-workspace",
        type=str,
        help="path to place a openvino workspace for nn-Meter builder"
    )
    platform.add_argument(
        "--customized-workspace",
        type=str,
        help="path to place a customized workspace for nn-Meter builder. A customized backend should be register first (refer to `nn-meter register --h` for more help)."
    )
    create_workspace.add_argument(
        "--backend",
        type=str,
        help="the backend name for registered backend"
    )
    create_workspace.set_defaults(func=create_workspace_cli)

    # Usage 4: test the connection to backend 
    # Usage: nn-meter connect --backend <backend-name> --workspace <path/to/workspace>
    test_connection = subparsers.add_parser(
        'connect', 
        help='connect to backend'
    )
    test_connection.add_argument(
        "--backend",
        type=str,
        help="the name of the testing backend"
    )
    test_connection.add_argument(
        "--workspace",
        type=str,
        help="path to the workspace with configuration completed"
    )
    test_connection.set_defaults(func=test_backend_connection_cli)
    
    # Usage 5: register customized module 
    # Usage: nn-meter register --backend <path/to/meta/file>
    register = subparsers.add_parser(
        'register', 
        help='register customized module to nn-Meter, supporting type: predictor, backend, operator, testcase, operator'
    )
    module_type = register.add_mutually_exclusive_group()
    module_type.add_argument(
        "--predictor",
        type=str,
        help="path to the meta file to register a customized predictor"
    )
    module_type.add_argument(
        "--backend",
        type=str,
        help="path to the meta file to register a customized backend"
    )
    module_type.add_argument(
        "--operator",
        type=str,
        help="path to the meta file to register a customized operator"
    )
    module_type.add_argument(
        "--testcase",
        type=str,
        help="path to the meta file to register a customized testcase"
    )
    module_type.add_argument(
        "--kernel",
        type=str,
        help="path to the meta file to register a customized kernel"
    )
    register.set_defaults(func=register_module_cli)
    
    # Usage 6: unregister customized module 
    # Usage: nn-meter unregister --backend <path/to/meta/file>
    unregister = subparsers.add_parser(
        'unregister', 
        help='unregister customized module from nn-Meter, supporting type: predictor, backend, operator, testcase, operator'
    )
    module_type = unregister.add_mutually_exclusive_group()
    module_type.add_argument(
        "--predictor",
        type=str,
        help="name of the registered predictor"
    )
    module_type.add_argument(
        "--backend",
        type=str,
        help="name of the registered backend"
    )
    module_type.add_argument(
        "--operator",
        type=str,
        nargs='+',
        help="name of the registered operator"
    )
    module_type.add_argument(
        "--testcase",
        type=str,
        nargs='+',
        help="name of the registered testcase"
    )
    module_type.add_argument(
        "--kernel",
        type=str,
        help="name of the registered kernel"
    )
    unregister.add_argument(
        "--predictor-version",
        type=float,
        help="the version of the predictor (if not specified, unregister all version)",
        default=None
    )
    unregister.set_defaults(func=unregister_module_cli)

    # Usage 7: change data folder
    # Usage: nn-meter set_data --data <path/to/new-folder>
    # TODO

    # parse args
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(stream=sys.stdout, format="(nn-Meter) %(message)s", level=logging.INFO)
    else:
        logging.basicConfig(stream=sys.stdout, format="(nn-Meter) %(message)s", level=logging.KEYINFO)
    args.func(args)


if __name__ == '__main__':
    nn_meter_cli()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import json
import requests
import numpy as np
from tqdm import tqdm
from zipfile import ZipFile


def download_from_url(urladdr, ppath):
    """
    download the kernel predictors from the url
    @params:

    urladdr: github release url address
    ppath: the targeting dir to save the download data (usually hardware_inferenceframework)

    """
    file_name = os.path.join(ppath, ".zip")
    if not os.path.isdir(ppath):
        os.makedirs(ppath)

    response = requests.get(urladdr, stream=True)
    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 2048  # 2 Kibibyte
    progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
    with open(file_name, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    zipfile = ZipFile(file_name)
    zipfile.extractall(path=ppath)
    zipfile.close() 
    progress_bar.close()
    os.remove(file_name)


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, (bytes, bytearray)):
            return obj.decode("utf-8")
        return json.JSONEncoder.default(self, obj)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import torch
import torch.nn as nn

def get_tensor_by_shapes(shapes, batch_size = 1):
    if len(shapes) == 1:
        return torch.randn(size=[batch_size] + shapes[0])
    else:
        return [torch.randn(size=[batch_size] + shape) for shape in shapes]


def get_inputs_by_shapes(shapes, batch_size = 1):
    return get_tensor_by_shapes(shapes, batch_size)


def get_padding(ks, s, hw):
    """ choose padding value to make sure:
    if s = 1, out_hw = in_hw;
    if s = 2, out_hw = in_hw // 2;
    if s = 4, out_hw = in_hw // 4;
    """
    if hw % s == 0:
        pad = max(ks - s, 0)
    else:
        pad = max(ks - (hw % s), 0)
    if pad % 2 == 0:
        return pad // 2
    else:
        return pad // 2 + 1



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import logging
import tensorflow as tf
import tensorflow.keras as keras
from .operators import *
from ..interface import BaseBlock
from .utils import get_inputs_by_shapes
logging = logging.getLogger("nn-Meter")


class TFBlock(BaseBlock):
    def __init__(self, config, batch_size = 1):
        self.config = config
        self.input_shape = [config["HW"], config["HW"], config["CIN"]]
        self.input_tensor_shape = [self.input_shape]
        self.batch_size = batch_size

    def test_block(self):
        import os, shutil
        from typing import List
        model_path = "./temp_model"
        model = self.get_model()
        model_output = model(get_inputs_by_shapes(self.input_tensor_shape))
        
        # check model save and reload
        keras.models.save_model(model, model_path)
        restore_model = keras.models.load_model(model_path)
        if isinstance(model_output, List):
            output_shape = [mod.shape for mod in model_output]
            restore_output_shape = [mod.shape for mod in restore_model(get_inputs_by_shapes(self.input_tensor_shape))]
        else:
            output_shape = model_output.shape
            restore_output_shape = restore_model(get_inputs_by_shapes(self.input_tensor_shape)).shape
        assert output_shape == restore_output_shape
        shutil.rmtree(model_path)

        # check model convert to tflite
        converter = tf.lite.TFLiteConverter.from_keras_model(restore_model)
        tflite_model = converter.convert()
        open(model_path + '.tflite', 'wb').write(tflite_model)
        os.remove(model_path + '.tflite')
        logging.keyinfo("Testing block is success!")

    def save_model(self, save_path):
        model = self.get_model()
        keras.models.save_model(model, save_path)

    def build_model(self, ops):
        ''' convert a list of operators to keras model.
        '''
        class Model(keras.Model):
            def __init__(self, ops):
                super().__init__()
                self.ops = keras.Sequential(ops)

            def call(self, inputs):
                x = self.ops(inputs)
                return x

        model = Model(ops)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model

    def get_model(self):
        raise NotImplementedError


class ConvBnRelu(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.bn_op, self.relu_op])


class ConvBnRelu6(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu6_op = Relu6(out_shape, config)
        self.relu6_op = relu6_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, conv_op, bn_op, relu6_op):
                super().__init__()
                self.conv = conv_op
                self.bn = bn_op
                self.relu6 = relu6_op

            def call(self, inputs):
                x = self.conv(inputs)
                x = self.bn(x)
                x = self.relu6(x)
                return x

        model = Model(self.conv_op, self.bn_op, self.relu6_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class ConvBn(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op = bn_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.bn_op])


class ConvRelu(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.relu_op])


class ConvRelu6(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()
        
        relu6_op = Relu6(out_shape, config)
        self.relu6_op = relu6_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, conv_op, relu6_op):
                super().__init__()
                self.conv = conv_op
                self.relu6 = relu6_op

            def call(self, inputs):
                x = self.conv(inputs)
                x = self.relu6(x)
                return x

        model = Model(self.conv_op, self.relu6_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class ConvHswish(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()
        
        hswish_op = Hswish(out_shape, config)
        self.hswish_op = hswish_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, conv_op, hswish_op):
                super().__init__()
                self.conv = conv_op
                self.hswish = hswish_op

            def call(self, inputs):
                x = self.conv(inputs)
                x = self.hswish(x)
                return x

        model = Model(self.conv_op, self.hswish_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class ConvBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op = conv_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op])


class ConvBnHswish(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        hswish_op = Hswish(out_shape, config)
        self.hswish_op = hswish_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, conv_op, bn_op, hswish_op):
                super().__init__()
                self.conv = conv_op
                self.bn = bn_op
                self.hswish = hswish_op

            def call(self, inputs):
                x = self.conv(inputs)
                x = self.bn(x)
                x = self.hswish(x)
                return x

        model = Model(self.conv_op, self.bn_op, self.hswish_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class ConvBnReluMaxPool(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op, out_shape = relu_op.get_model(), relu_op.get_output_shape()
        
        maxpool_op = MaxPool(out_shape, config)
        self.maxpool_op = maxpool_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.bn_op, self.relu_op, self.maxpool_op])


class DwConvBn(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op = bn_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op, self.bn_op])


class DwConvRelu(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op, self.relu_op])


class DwConvRelu6(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()
        
        relu6_op = Relu6(out_shape, config)
        self.relu6_op = relu6_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, dwconv_op, relu6_op):
                super().__init__()
                self.dwconv = dwconv_op
                self.relu6 = relu6_op

            def call(self, inputs):
                x = self.dwconv(inputs)
                x = self.relu6(x)
                return x

        model = Model(self.dwconv_op, self.relu6_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class DwConvBnRelu(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op, self.bn_op, self.relu_op])


class DwConvBnRelu6(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu6_op = Relu6(out_shape, config)
        self.relu6_op = relu6_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, dwconv_op, bn_op, relu6_op):
                super().__init__()
                self.dwconv = dwconv_op
                self.bn = bn_op
                self.relu6 = relu6_op

            def call(self, inputs):
                x = self.dwconv(inputs)
                x = self.bn(x)
                x = self.relu6(x)
                return x

        model = Model(self.dwconv_op, self.bn_op, self.relu6_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class DwConvBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op = dwconv_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op])


class ConvBnHswish(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        hswish_op = Hswish(out_shape, config)
        self.hswish_op = hswish_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, dwconv_op, bn_op, hswish_op):
                super().__init__()
                self.dwconv = dwconv_op
                self.bn = bn_op
                self.hswish = hswish_op

            def call(self, inputs):
                x = self.dwconv(inputs)
                x = self.bn(x)
                x = self.hswish(x)
                return x

        model = Model(self.dwconv_op, self.bn_op, self.hswish_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class MaxPoolBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        maxpool_op = MaxPool(self.input_shape, config)
        self.maxpool_op = maxpool_op.get_model()

    def get_model(self):
        return self.build_model([self.maxpool_op])


class AvgPoolBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        avgpool_op = AvgPool(self.input_shape, config)
        self.avgpool_op = avgpool_op.get_model()

    def get_model(self):
        return self.build_model([self.avgpool_op])


class FCBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        self.config = config
        self.input_shape = [config["CIN"]]
        self.input_tensor_shape = [self.input_shape]
        self.batch_size = batch_size

        fc_op = FC(self.input_shape, config)
        self.fc_op = fc_op.get_model()

    def get_model(self):
        return self.build_model([self.fc_op])


class ConcatBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        self.config = config
        self.input_shape = [[config["HW"], config["HW"], cin]
                       for cin in [config['CIN1'], config['CIN2'], config['CIN3'], config['CIN4']]
                       if cin != 0]
        self.input_tensor_shape = self.input_shape
        self.batch_size = batch_size
        
        concat_op = Concat(self.input_shape, config)
        self.concat_op = concat_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, concat_op):
                super().__init__()
                self.concat = concat_op

            def call(self, inputs):
                return self.concat(inputs)

        model = Model(self.concat_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class SplitBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        split_op = Split(self.input_shape, config)
        self.split_op = split_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, split_op):
                super().__init__()
                self.split = split_op

            def call(self, inputs):
                return self.split(inputs)

        model = Model(self.split_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class ChannelShuffle(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

    def get_model(self):
        class Model(keras.Model):
            def __init__(self):
                super().__init__()

            def call(self, inputs):
                _, h, w, c = inputs.get_shape().as_list()
                x = tf.reshape(inputs, [-1, h, w, 2, c // 2])
                x = tf.transpose(x, (0, 1, 2, 4, 3))
                x = tf.reshape(x, [-1, h, w, c])
                return x

        model = Model()
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class SEBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        se_op = SE(self.input_shape, config)
        self.se_op = se_op.get_model()

    def get_model(self):
        return self.build_model([self.se_op])


class GlobalAvgPoolBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        globalavgpool_op = GlobalAvgpool(self.input_shape, config)
        self.globalavgpool_op = globalavgpool_op.get_model()

    def get_model(self):
        return self.build_model([self.globalavgpool_op])


class BnRelu(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        bn_op = BN(self.input_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.bn_op, self.relu_op])


class BnBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        bn_op = BN(self.input_shape, config)
        self.bn_op = bn_op.get_model()

    def get_model(self):
        return self.build_model([self.bn_op])


class HswishBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        hswish_op = Hswish(self.input_shape, config)
        self.hswish_op = hswish_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, hswish_op):
                super().__init__()
                self.hswish = hswish_op

            def call(self, inputs):
                return self.hswish(inputs)

        model = Model(self.hswish_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class ReluBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        relu_op = Relu(self.input_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.relu_op])


class AddRelu(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        add_op = Add(self.input_shape, config)
        self.add_op, out_shape = add_op.get_model(), add_op.get_output_shape()

        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, add_op, relu_op):
                super().__init__()
                self.add = add_op
                self.relu = relu_op

            def call(self, inputs):
                x = self.add([inputs, inputs])
                x = self.relu(x)
                return x

        model = Model(self.add_op, self.relu_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class AddBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)
        
        add_op = Add(self.input_shape, config)
        self.add_op = add_op.get_model()

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, add_op):
                super().__init__()
                self.add = add_op

            def call(self, inputs):
                return self.add([inputs, inputs])

        model = Model(self.add_op)
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class GroupedConvBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)
        self.cout = self.input_shape[2] if "COUT" not in config else config["COUT"]
        self.num_groups = config['NUM_GROUPS']

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, cout, num_groups, kernel_size, strides):
                super().__init__()
                self.cout = cout
                self.num_groups = num_groups
                self.kernel_size = kernel_size
                self.strides = strides

            def call(self, inputs):
                x = [keras.layers.Conv2D(
                        filters=self.cout // self.num_groups,
                        kernel_size=self.kernel_size,
                        strides=self.strides,
                        padding="same",
                    )(x) for x in tf.split(inputs, self.num_groups, axis=3)
                ]
                return tf.concat(x, axis=3)    

        model = Model(self.cout, self.num_groups, self.config['KERNEL_SIZE'], self.config['STRIDES'])
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model


class MixedConvBlock(TFBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)
        self.cout = self.input_shape[2] if "COUT" not in config else config["COUT"]
        self.num_groups = config['NUM_GROUPS']

    def get_model(self):
        class Model(keras.Model):
            def __init__(self, cout, num_groups, strides):
                super().__init__()
                self.cout = cout
                self.num_groups = num_groups
                self.strides = strides

            def call(self, inputs):
                x = [keras.layers.Conv2D(
                        filters=self.cout // self.num_groups,
                        kernel_size=i * 2 + 3,
                        strides=self.strides,
                        padding="same",
                    )(x) for i, x in zip(range(self.num_groups), tf.split(inputs, self.num_groups, axis=3))
                ]
                return tf.concat(x, axis=3)    

        model = Model(self.cout, self.num_groups, self.config['STRIDES'])
        model(get_inputs_by_shapes(self.input_tensor_shape, self.batch_size))
        return model



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

try:
    import pkg_resources  # part of setuptools
    __version__ = pkg_resources.require("nn-meter")[0].version
except ModuleNotFoundError:
    __version__ = 'UNKNOWN'

from functools import partial, partialmethod

from .predictor import (
    nnMeterPredictor,
    load_latency_predictor,
    list_latency_predictors,
    latency_metrics
)
from .ir_converter import (
    model_file_to_graph,
    model_to_graph
)
from .utils import (
    create_user_configs,
    change_user_data_folder,
    download_from_url
)
from .dataset import bench_dataset

# setup logging
import sys
import logging
logging.KEYINFO = 22
logging.addLevelName(logging.KEYINFO, 'KEYINFO')
logging.Logger.keyinfo = partialmethod(logging.Logger.log, logging.KEYINFO)
logging.keyinfo = partial(logging.log, logging.KEYINFO)

logging.RESULT = 25
logging.addLevelName(logging.RESULT, 'RESULT')
logging.Logger.result = partialmethod(logging.Logger.log, logging.RESULT)
logging.result = partial(logging.log, logging.RESULT)

from logging import Formatter, StreamHandler
handler = StreamHandler(sys.stdout)
handler.setFormatter(Formatter("(nn-Meter) %(message)s"))
logger = logging.getLogger("nn-Meter")
logger.addHandler(handler)
logger.setLevel(logging.INFO)
logger.propagate = False


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import re
from .openvino_profiler import OpenVINOProfiler
from .openvino_backend import OpenVINOBackend
from ..interface import BaseParser
from nn_meter.builder.backend_meta.utils import Latency, ProfiledResults


class OpenVINOVPULatencyParser(BaseParser):

    def parse(self, content):
        self.layers = self._parse_layers(content)
        self.comp_layer_latency = sum(
            Latency(layer['realtime'])
            for layer in self.layers
            if layer['layer_name'] != '<Extra>'
        )
        return self

    def _parse_layers(self, content):
        layer_regex = r'^([^;]+);([^;]+);([^;]+);([^;]+);([^;]+);([^;]+);$'
        layers = []
        for match in re.findall(layer_regex, content, re.MULTILINE):
            try:
                layers.append({
                    'layer_name': match[0],
                    'exec_status': match[1],
                    'layer_type': match[2],
                    'exec_type': match[3],
                    'realtime': float(match[4]),
                    'cputime': float(match[5]),
                })
            except:
                pass
        return layers

    @property
    def latency(self):
        return self.comp_layer_latency
    
    @property
    def results(self):
        results = ProfiledResults({'latency': self.latency})
        return results


class OpenVINOVPUProfiler(OpenVINOProfiler):
    device = "MYRIAD"


class OpenVINOVPUBackend(OpenVINOBackend):
    parser_class = OpenVINOVPULatencyParser
    profiler_class = OpenVINOVPUProfiler



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
from ..interface import BaseProfiler


class TFLiteProfiler(BaseProfiler):
    use_gpu = None

    def __init__(self, dst_kernel_path, benchmark_model_path, graph_path='', dst_graph_path='', serial='', num_threads=1, num_runs=50, warm_ups=10):
        """
        @params:
        graph_path: graph file. path on host server
        dst_graph_path: graph file. path on android device
        kernel_path: dest kernel output file. path on android device
        benchmark_model_path: path to benchmark_model on android device
        """
        self._serial = serial
        self._graph_path = graph_path
        self._dst_graph_path = dst_graph_path
        self._dst_kernel_path = dst_kernel_path
        self._benchmark_model_path = benchmark_model_path
        self._num_threads = num_threads
        self._num_runs = num_runs
        self._warm_ups = warm_ups

    def profile(self, graph_path, preserve = False, clean = True, taskset = '70', close_xnnpack = False):
        """
        @params:
        preserve: tflite file exists in remote dir. No need to push it again.
        clean: remove tflite file after running.
        """
        model_name = os.path.basename(graph_path)
        remote_graph_path = os.path.join(self._dst_graph_path, model_name)
        taskset_cmd = f'taskset {taskset}' if taskset else ''
        kernel_cmd = f'--kernel_path={self._dst_kernel_path}' if self._dst_kernel_path else ''
        close_xnnpack_cmd = f'--use_xnnpack=false' if close_xnnpack else ''

        from ppadb.client import Client as AdbClient
        client = AdbClient(host="127.0.0.1", port=5037)
        if self._serial:
            device = client.device(self._serial)
        else:
            device = client.devices()[0]

        if not preserve:
            device.push(graph_path, remote_graph_path)
        try:
            taskset_cmd = f'taskset {taskset}' if taskset else '' 
            kernel_cmd = f'--kernel_path={self._dst_kernel_path}' if self._dst_kernel_path else ''
            close_xnnpack_cmd = f'--use_xnnpack=false' if close_xnnpack else ''
            res = device.shell(f' {taskset_cmd} {self._benchmark_model_path} {kernel_cmd} {close_xnnpack_cmd}' \
                               f' {close_xnnpack_cmd}' \
                               f' --num_threads={self._num_threads}' \
                               f' --num_runs={self._num_runs}' \
                               f' --warmup_runs={self._warm_ups}' \
                               f' --graph={remote_graph_path}' \
                               f' --enable_op_profiling=true' \
                               f' --use_gpu={"true" if self.use_gpu else "false"}')
        except:
            raise
        finally:
            if clean:
                if self._serial:
                    os.system(f"adb -s {self._serial} shell rm {remote_graph_path}")
                    os.remove(graph_path)
                else:
                    os.system(f"adb shell rm {remote_graph_path}")

        return res



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import sys
import yaml
import logging
import importlib
from . import config_sampler
logging = logging.getLogger("nn-Meter")


__BUILTIN_KERNELS__ = {
    # builtin name: [kernel class name, kernel sampler class name]
    "conv-bn-relu": ["ConvBnRelu", "ConvSampler"],
    "conv-bn-relu6": ["ConvBnRelu6", "ConvSampler"],
    "conv-bn": ["ConvBn", "ConvSampler"],
    "conv-relu": ["ConvRelu", "ConvSampler"],
    "conv-relu6": ["ConvRelu6", "ConvSampler"],
    "conv-hswish": ["ConvHswish", "ConvSampler"],
    "conv-block": ["ConvBlock", "ConvSampler"],
    "conv-bn-hswish": ["ConvBnHswish", "ConvSampler"],
    # dwconv
    "dwconv-bn": ["DwConvBn", "DwConvSampler"],
    "dwconv-relu": ["DwConvRelu", "DwConvSampler"],
    "dwconv-relu6": ["DwConvRelu6", "DwConvSampler"],
    "dwconv-bn-relu": ["DwConvBnRelu", "DwConvSampler"],
    "dwconv-bn-relu6": ["DwConvBnRelu6", "DwConvSampler"],
    "dwconv-block": ["DwConvBlock", "DwConvSampler"],
    "dwconv-bn-hswish": ["ConvBnHswish", "DwConvSampler"],
    # others
    "maxpool": ["MaxPoolBlock", "PoolingSampler"],
    "avgpool": ["AvgPoolBlock", "PoolingSampler"],
    "fc": ["FCBlock", "FCSampler"],
    "concat": ["ConcatBlock", "ConcatSampler"],
    "split": ["SplitBlock", "CinEvenSampler"],
    "channelshuffle": ["ChannelShuffle", "CinEvenSampler"],
    "se": ["SEBlock", "CinEvenSampler"],
    "global-avgpool": ["GlobalAvgPoolBlock", "GlobalAvgPoolSampler"],
    "bnrelu": ["BnRelu", "HwCinSampler"],
    "bn": ["BnBlock", "HwCinSampler"],
    "hswish": ["HswishBlock", "HwCinSampler"],
    "relu": ["ReluBlock", "HwCinSampler"],
    "addrelu": ["AddRelu", "HwCinSampler"],
    "add": ["AddBlock", "HwCinSampler"],
}


__user_config_folder__ = os.path.expanduser('~/.nn_meter/config')
__registry_cfg_filename__ = 'registry.yaml'
__REG_KERNELS__ = {}
if os.path.isfile(os.path.join(__user_config_folder__, __registry_cfg_filename__)):
    with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'r') as fp:
        registry_modules = yaml.load(fp, yaml.FullLoader)
    if "kernels" in registry_modules:
        __REG_KERNELS__ = registry_modules["kernels"]


def generate_model_for_kernel(kernel_type, config, save_path, implement='tensorflow', batch_size=1):
    """ get the nn model for predictor build.
    """
    # get kernel class information
    if kernel_type in __REG_KERNELS__:
        kernel_info = __REG_KERNELS__[kernel_type]
        assert kernel_info["implement"] == implement
        sys.path.append(kernel_info["package_location"])
        kernel_name = kernel_info["class_name"]
        kernel_module = importlib.import_module(kernel_info["class_module"])

    elif kernel_type in __BUILTIN_KERNELS__:
        kernel_name = __BUILTIN_KERNELS__[kernel_type][0]
        if implement == 'tensorflow':
            from nn_meter.builder.nn_modules.tf_networks import blocks
        elif implement == 'torch':
            from nn_meter.builder.nn_modules.torch_networks import blocks
        else:
            raise NotImplementedError('You must choose one implementation of kernel from "tensorflow" or "pytorch"')
        kernel_module = blocks

    else:
        raise ValueError(f"Unsupported kernel type: {kernel_type}. Please register the kernel first.")

    # get kernel class and create kernel instance by needed_config
    kernel_class = getattr(kernel_module, kernel_name)(config, batch_size)
    input_tensor_shape = kernel_class.input_tensor_shape
    model = kernel_class.get_model()

    # save model file to savepath
    kernel_class.save_model(save_path)
    logging.info(f"{kernel_type} model is generated and saved to {save_path}.")

    return model, input_tensor_shape, config


def get_sampler_for_kernel(kernel_type, sample_num, sampling_mode, configs = None):
    """ return the list of sampled data configurations in prior and finegrained sampling mode
    """
    # get kernel sampler class information
    if kernel_type in __REG_KERNELS__:
        kernel_info = __REG_KERNELS__[kernel_type]
        sys.path.append(kernel_info["package_location"])
        sampler_name = kernel_info["sampler_name"]
        sampler_module = importlib.import_module(kernel_info["sampler_module"])
    elif kernel_type in __BUILTIN_KERNELS__:
        sampler_name = __BUILTIN_KERNELS__[kernel_type][1]
        sampler_module = config_sampler

    # get kernel class and create kernel instance by needed_config
    sampler_class = getattr(sampler_module, sampler_name)()

    # initialize sampling, based on prior distribution
    if sampling_mode == 'prior':
        sampled_cfgs = sampler_class.prior_config_sampling(sample_num)
    # fine-grained sampling for data with large error points
    elif sampling_mode == 'finegrained':
        sampled_cfgs = sampler_class.finegrained_config_sampling(configs, sample_num)
    return sampled_cfgs


def list_kernels():
    return list(__BUILTIN_KERNELS__.keys()) + ["* " + item for item in list(__REG_KERNELS__.keys())]



import os
import pandas as pd

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def read_conv_zoo(filename = "conv.csv"):
    filename = os.path.join(BASE_DIR, filename)
    conv_df = pd.read_csv(filename)
    hws = conv_df["input_h"]
    cins = conv_df["cin"]
    couts = conv_df["cout"]
    ks = conv_df["ks"]
    strides = conv_df["stride"]
    return hws, cins, couts, ks, strides


def read_dwconv_zoo(filename = "dwconv.csv"):
    filename = os.path.join(BASE_DIR, filename)
    dwconv_df = pd.read_csv(filename)
    hws = dwconv_df["input_h"]
    cins = dwconv_df["cin"]
    ks = dwconv_df["ks"]
    strides = dwconv_df["stride"]
    return hws, cins, ks, strides


def read_fc_zoo(filename = "fc.csv"):
    filename = os.path.join(BASE_DIR, filename)
    fc_df = pd.read_csv(filename)
    cins = fc_df["cin"]
    couts = fc_df["cout"]
    return cins, couts


def read_pool_zoo(filename = "pooling.csv"):
    filename = os.path.join(BASE_DIR, filename)
    pool_df = pd.read_csv(filename)
    hws = pool_df["input_h"]
    cins = pool_df["cin"]
    ks = pool_df["ks"]
    strides = pool_df["stride"]
    return hws, cins, ks, strides



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
nni_type_map = {
    "aten::mul": "mul",
    "aten::floordiv": "div",
    "aten::reshape": "reshape",
    "aten::cat": "concat",
    "aten::split": "split",
    "__torch__.torch.nn.modules.conv.Conv2d": "conv",
    "__torch__.torch.nn.modules.activation.ReLU": "relu",
    "__torch__.torch.nn.modules.batchnorm.BatchNorm2d": "bn",
    "__torch__.torch.nn.modules.linear.Linear": "fc",
    "__torch__.torch.nn.modules.pooling.AvgPool2d": "gap",
    "__torch__.torch.nn.modules.pooling.MaxPool2d": "maxpool",
    "__torch__.torch.nn.modules.activation.Sigmoid": "sigmoid",
    "__torch__.torch.nn.modules.activation.Hardsigmoid": "hardsigmoid",
    "__torch__.torch.nn.modules.activation.Hardswish": "hswish",
    "__torch__.torch.nn.modules.activation.ReLU6": "relu",
    "__torch__.torch.nn.modules.activation.Softmax": "softmax",
}


def int_to_list_modifier(attr):
    if isinstance(attr, int):
        return [attr, attr]
    else:
        return list(attr)


nni_attr_map = {
    "__all__": {
        "kernel_size": ("ks", int_to_list_modifier),
        "padding": ("pads", int_to_list_modifier),
        "stride": ("strides", int_to_list_modifier),
        "groups": ("group", None),
    },
    "concat": {
        "dim": ("axis", None),
    },
}



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import subprocess
import numpy as np
import shutil
import serial

from .utils import restart
from ..interface import BaseProfiler
from nn_meter.utils.pyutils import get_pyver


class OpenVINOProfiler(BaseProfiler):

    device = None

    def __init__(self, venv, optimizer, runtime_dir, serial, graph_path='', _dst_graph_path='', data_type='FP16'):
        self._graph_path = graph_path
        self._venv = venv
        self._optimizer = optimizer
        self._dst_graph_path = _dst_graph_path
        self._runtime_dir = runtime_dir
        self._serial = serial
        self._data_type = data_type

    def load_graph(self, graph_path, dst_graph_path):
        self._graph_path = graph_path
        self._dst_graph_path = dst_graph_path

    def profile(self, shapes, retry=2):
        interpreter_path = os.path.join(self._venv, 'bin/python')
        pyver = get_pyver(interpreter_path)

        subprocess.run(
            f'{interpreter_path} {self._optimizer} '
            f'--input_model {self._graph_path} '
            f'--output_dir {self._dst_graph_path} '
            f'--data_type {self._data_type}',
            shell=True
        )

        filename = os.path.splitext(os.path.basename(self._graph_path))[0]

        input_path = os.path.join(self._dst_graph_path, 'inputs')
        if os.path.exists(input_path):
            shutil.rmtree(input_path)

        os.mkdir(input_path)
        for index, shape in enumerate(shapes):
            np.random.rand(*shape).astype('float32').tofile(os.path.join(input_path, f'input_{index}.bin'))

        output = ''

        with serial.Serial(self._serial, 115200, timeout=1) as ser:
            restart(ser)

            command = (
                f'. {os.path.join(self._venv, "bin/activate")}; '
                f'cd {self._runtime_dir}; '
                f'. {os.path.join(self._runtime_dir, "setupvars.sh")} -pyver {pyver}; '
                f'{os.path.join(self._runtime_dir, "benchmark_app")} '
                f'-i {input_path} '
                f'-m {os.path.join(self._dst_graph_path, filename + ".xml")} '
                f'-d {self.device} '
                f'-report_type detailed_counters '
                f'-report_folder {self._dst_graph_path} '
                f'-niter 50 '
                f'-nireq 1 '
                f'-api sync'
            )

            while True:
                try:
                    subprocess.run(
                        f'bash -c "{command}"',
                        shell=True,
                        timeout=30,
                    )
                    output = open(os.path.join(self._dst_graph_path, 'benchmark_detailed_counters_report.csv'), 'r').read()
                    break
                except subprocess.TimeoutExpired as e:
                    print(e)

                    if retry == 0:
                        raise e
                    print('Retrying...')
                    restart(ser)

                    retry -= 1

        return output



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import pandas as pd
import logging
from sklearn.model_selection import train_test_split
from .utils import collect_kernel_data, latency_metrics
from .predictor_lib import init_predictor
from .extract_feature import get_feature_parser, get_data_by_profiled_results
logging = logging.getLogger("nn-Meter")


def build_predictor_by_data(kernel_type, kernel_data, backend = None, error_threshold = 0.1, mark = '', save_path = None, predict_label = "latency", final_predictor=False):
    """
    build regression model by sampled data and latency, locate data with large-errors. Returns (current predictor, 10% Accuracy, error_cfgs), 
    where error_cfgs represent configuration list, where each item is a configuration for one large-error-data.

    @params
    kernel_type (str): type of target kernel

    data (tuple): feature (configs) and target (latencies) data

    backend (str): target device, relative to predictor initialization
    
    error_threshold (float): default = 0.1, should be no less than 0.1. if prediction error (`abs(pred - true) / true`) > error_threshold,
        we treat this data as a large-error-data.

    mark (str): the mark for the running results. Defaults to ''.

    save_path (str): the folder to save results file such as feature table and predictor pkl file. If save_path is None, the data will not be saved.
    
    predict_label (str): the predicting label to build kernel predictor
    """
    feature_parser = get_feature_parser(kernel_type)
    if save_path:
        os.makedirs(os.path.join(save_path, "collection"), exist_ok=True)
        os.makedirs(os.path.join(save_path, "predictors"), exist_ok=True)
        data_save_path = os.path.join(save_path, "collection", f'Data_{kernel_type}_{mark}.csv')
        res_save_path = os.path.join(save_path, "collection", f"TestResult_{kernel_type}_{mark}.csv")
        pred_save_path = os.path.join(save_path, "predictors", f"{kernel_type}_{mark}.pkl")
    else:
        data_save_path = None
        res_save_path = None
        pred_save_path = None

    kernel_data = collect_kernel_data(kernel_data, predict_label)
    data = get_data_by_profiled_results(kernel_type, feature_parser, kernel_data,
                                        save_path=data_save_path,
                                        predict_label=predict_label)

    acc10, error_configs = None, None
    X, Y = data
    # initialize the regression model based on `RandomForestRegressor`
    predictor = init_predictor(kernel_type, backend)

    if final_predictor:
        predictor.fit(X, Y)
    else:
        # get data for regression
        trainx, testx, trainy, testy = train_test_split(X, Y, test_size = 0.2, random_state = 10)
        logging.info(f"training data size: {len(trainx)}, test data size: {len(testx)}")

        # start training
        predictor.fit(trainx, trainy)
        predicts = predictor.predict(testx)
        pred_error_list = [abs(y1 - y2) / y1 for y1, y2 in zip(testy, predicts)]
        rmse, rmspe, error, acc5, acc10, acc15 = latency_metrics(predicts, testy)
        logging.info(f"rmse: {rmse:.4f}; rmspe: {rmspe:.4f}; error: {error:.4f}; 5% accuracy: {acc5:.4f}; 10% accuracy: {acc10:.4f}; 15% accuracy: {acc15:.4f}.")
        
        # dump the test set with predicts to csv file
        test_res = pd.DataFrame(testx, columns=[f'feature{i}' for i in range(len(testx[0]))])
        test_res["True"] = testy
        test_res["Pred"] = predicts
        test_res["Error"] = pred_error_list
        if res_save_path:
            test_res.to_csv(res_save_path, index=False)
            logging.info(f"All test data and predicted results are stored in path {res_save_path}")

        # locate large error data
        error_configs = []
        for i in range(len(testx)):
            if pred_error_list[i] > error_threshold:
                error_config = feature_parser.get_config_by_feature(testx[i])
                error_configs.append(error_config)

    # dump the predictor model
    if pred_save_path:
        import pickle
        with open(pred_save_path, 'wb') as fp:
            pickle.dump(predictor, fp)
        logging.keyinfo(f"Saved the predictor for {kernel_type} in path {pred_save_path}.")

    return predictor, acc10, error_configs



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .utils import get_kernel_name
from .extract_feature import get_predict_features


def merge_conv_kernels(kernelname):
    """
    to speed up, we merge conv and dwconv related kernels into one kernel by their name
    """
    if "conv" in kernelname and "dwconv" not in kernelname:
        return "conv-bn-relu"
    elif "dwconv" in kernelname:
        return "dwconv-bn-relu"
    else:
        return kernelname


def predict_model(model, predictors):
    """
    @params:
    model: the model config with prediction features
    predictors: loaded pkl predictors
    """
    py = 0
    dicts = {}
    for layer in model:
        kernel = list(model[layer].keys())[0]
        features = model[layer][kernel]
        rkernel = merge_conv_kernels(kernel)
        if rkernel not in dicts:
            dicts[rkernel] = []
        dicts[rkernel].append(features)

    for kernel in dicts:
        kernelname = get_kernel_name(kernel)
        if kernelname in predictors:
            pred = predictors[kernelname]
            pys = pred.predict(dicts[kernel]) # in unit of ms
            if len(pys) != 0:
                py += sum(pys)

    return py


def nn_predict(predictors, kernel_units):
    """
    @params:
    predictors: dictionary object, key: kernel name, object: loaded pkl latency model
    kernel_units: the divided kernel units and the features of a model.
    """

    features = get_predict_features(kernel_units)
    py = predict_model(features, predictors)
    return py



import subprocess


def get_pyver(interpreter_path):
    script = '''
import sys
print(f'{sys.version_info.major}.{sys.version_info.minor}')
    '''
    version = subprocess.check_output(f'{interpreter_path} -c "{script}"', shell=True)
    return version.decode('utf-8').strip()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from tensorflow import keras
from .interface import BaseTestCase


class SingleOpModel(keras.Model):
    def __init__(self, op):
        super().__init__()
        self.op = op

    def call(self, inputs):
        return self.op(inputs)


class TwoOpModel(keras.Model):
    def __init__(self, op1, op2, op1_is_two_inputs, op2_is_two_inputs):
        super().__init__()
        self.op1 = op1
        self.op2 = op2
        self.op1_is_two_inputs = op1_is_two_inputs
        self.op2_is_two_inputs = op2_is_two_inputs

    def call(self, inputs):
        if self.op1_is_two_inputs:
            x = self.op1([inputs[0], inputs[1]])
        else:
            if self.op2_is_two_inputs:
                x = self.op1(inputs[0])
            else:
                x = self.op1(inputs)
        if self.op2_is_two_inputs:
            x = self.op2([x, inputs[-1]])
        else:
            x = self.op2(x)
        return x


class MultipleOutNodes(BaseTestCase):
    name = 'MON'
    cases = {
        'case1': ['relu_relu', 'relu_dwconv', 'dwconv'],
        'case2': ['dwconv_relu_relu', 'relu_dwconv'],
        'case3': ['dwconv_relu', 'dwconv', 'relu_relu']
    }
    true_case = 'case1'
    deps = {
        'BF_dwconv_relu': True,
    }
    implement = 'tensorflow'

    def _model_block(self):
        input_layer = keras.Input(shape=self.input_shape)

        x = keras.layers.DepthwiseConv2D(self.kernel_size, padding=self.padding)(input_layer)
        branch_1 = keras.layers.ReLU(negative_slope=0)(x)
        branch_1 = keras.layers.ReLU(negative_slope=0)(branch_1)
        branch_2 = keras.layers.ReLU(negative_slope=2)(x)
        branch_2 = keras.layers.DepthwiseConv2D(self.kernel_size, padding=self.padding)(branch_2)

        return keras.models.Model(input_layer, [branch_1, branch_2]), [self.input_shape]

    def _model_relu_relu(self):
        input_layer = keras.Input(shape=self.input_shape)

        x = keras.layers.ReLU()(input_layer)
        x = keras.layers.ReLU()(x)

        return keras.models.Model(input_layer, x), [self.input_shape]

    def _model_dwconv_relu_relu(self):
        input_layer = keras.Input(shape=self.input_shape)

        x = keras.layers.DepthwiseConv2D(self.kernel_size, padding=self.padding)(input_layer)
        x = keras.layers.ReLU()(x)
        x = keras.layers.ReLU()(x)

        return keras.models.Model(input_layer, x), [self.input_shape]

    def _model_relu_dwconv(self):
        input_layer = keras.Input(shape=self.input_shape)

        x = keras.layers.ReLU()(input_layer)
        x = keras.layers.DepthwiseConv2D(self.kernel_size, padding=self.padding)(x)

        return keras.models.Model(input_layer, x), [self.input_shape]

    def _model_dwconv_relu(self):
        input_layer = keras.Input(shape=self.input_shape)

        x = keras.layers.DepthwiseConv2D(self.kernel_size, padding=self.padding)(input_layer)
        x = keras.layers.ReLU()(x)

        return keras.models.Model(input_layer, x), [self.input_shape]

    def _model_dwconv(self):
        input_layer = keras.Input(shape=self.input_shape)

        x = keras.layers.DepthwiseConv2D(self.kernel_size, padding=self.padding)(input_layer)

        return keras.models.Model(input_layer, x), [self.input_shape]




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import torch
import numpy as np
import torch.nn as nn
from .utils import get_padding
from ..interface import BaseOperator

'''
This file contains the torch implementation of operators
'''

#---------------------- convolution layer ----------------------#

class Conv(BaseOperator):
    def get_model(self):
        cin = self.input_shape[0]
        cout = cin if "COUT" not in self.config else self.config["COUT"]
        padding = get_padding(self.config["KERNEL_SIZE"], self.config["STRIDES"], self.input_shape[1])
        return nn.Conv2d(cin, cout, kernel_size=self.config["KERNEL_SIZE"], stride=self.config["STRIDES"], padding=padding)

    def get_output_shape(self):
        cout = self.input_shape[0] if "COUT" not in self.config else self.config["COUT"]
        output_h = (self.input_shape[1] - 1) // self.config["STRIDES"] + 1
        output_w = (self.input_shape[2] - 1) // self.config["STRIDES"] + 1
        return [cout, output_h, output_w]


class DwConv(BaseOperator):
    def get_model(self):
        cin = self.input_shape[0]
        padding = get_padding(self.config["KERNEL_SIZE"], self.config["STRIDES"], self.input_shape[1])
        return nn.Conv2d(cin, cin, kernel_size=self.config["KERNEL_SIZE"], stride=self.config["STRIDES"], padding=padding, groups=cin)

    def get_output_shape(self):
        cin = self.input_shape[0]
        output_h = (self.input_shape[1] - 1) // self.config["STRIDES"] + 1
        output_w = (self.input_shape[2] - 1) // self.config["STRIDES"] + 1
        return [cin, output_h, output_w]


class ConvTrans(BaseOperator):
    def get_model(self):
        cin = self.input_shape[0]
        cout = cin if "COUT" not in self.config else self.config["COUT"]
        padding = get_padding(self.config["KERNEL_SIZE"], self.config["STRIDES"], self.input_shape[1])
        output_padding = self.config["STRIDES"] + 2 * padding - self.config["KERNEL_SIZE"]
        return nn.ConvTranspose2d(cin, cout, kernel_size=self.config["KERNEL_SIZE"], stride=self.config["STRIDES"], padding=padding, output_padding=output_padding)

    def get_output_shape(self):
        cout = self.input_shape[0] if "COUT" not in self.config else self.config["COUT"]
        return [cout, self.input_shape[1] * self.config["STRIDES"], self.input_shape[2] * self.config["STRIDES"]]

#------------------ normalization and pooling ------------------#

class BN(BaseOperator):
    def get_model(self):
        cin = self.input_shape[0]
        return nn.BatchNorm2d(cin)


class GlobalAvgpool(BaseOperator):
    def get_model(self):
        return nn.AdaptiveAvgPool2d([1, 1])
    
    def get_output_shape(self):
        cin = self.input_shape[0]
        return [cin, 1, 1]


class MaxPool(BaseOperator):
    def get_model(self):
        if "POOL_STRIDES" not in self.config:
            self.config["POOL_STRIDES"] = self.config["STRIDES"]
        padding = get_padding(self.config["KERNEL_SIZE"], self.config["POOL_STRIDES"], self.input_shape[1])
        return nn.MaxPool2d(self.config["KERNEL_SIZE"], self.config["POOL_STRIDES"], padding=padding)

    def get_output_shape(self):
        cin = self.input_shape[0]
        if "POOL_STRIDES" not in self.config:
            self.config["POOL_STRIDES"] = self.config["STRIDES"]
        output_h = (self.input_shape[1] - 1) // self.config["POOL_STRIDES"] + 1
        output_w = (self.input_shape[2] - 1) // self.config["POOL_STRIDES"] + 1
        return [cin, output_h, output_w]


class AvgPool(BaseOperator):
    def get_model(self):
        if "POOL_STRIDES" not in self.config:
            self.config["POOL_STRIDES"] = self.config["STRIDES"]
        padding = get_padding(self.config["KERNEL_SIZE"], self.config["POOL_STRIDES"], self.input_shape[1])
        return nn.AvgPool2d(self.config["KERNEL_SIZE"], self.config["POOL_STRIDES"], padding=padding)

    def get_output_shape(self):
        cin = self.input_shape[0]
        if "POOL_STRIDES" not in self.config:
            self.config["POOL_STRIDES"] = self.config["STRIDES"]
        output_h = (self.input_shape[1] - 1) // self.config["POOL_STRIDES"] + 1
        output_w = (self.input_shape[2] - 1) // self.config["POOL_STRIDES"] + 1
        return [cin, output_h, output_w]

#------------------------ other modules ------------------------#

class SE(BaseOperator):
    def get_model(self):
        class SE(nn.Module):
            def __init__(self, num_channels, se_ratio=0.25):
                super().__init__()
                mid_channels = int(num_channels * se_ratio)
                self.squeeze = nn.Conv2d(num_channels, mid_channels, kernel_size=1, padding=0)
                self.relu = nn.ReLU()
                self.excite = nn.Conv2d(mid_channels, num_channels, kernel_size=1, padding=0)
                self.hswish = nn.Hardswish()

            def _scale(self, x):
                x = x.mean(3, keepdim=True).mean(2, keepdim=True)
                x = self.squeeze(x)
                x = self.relu(x)
                x = self.excite(x)
                x = self.hswish(x)
                return x

            def forward(self, x):
                scale = self._scale(x)
                return scale * x
        return SE(self.input_shape[0])


class FC(BaseOperator):
    def get_model(self):
        cin = self.input_shape[0]
        cout = self.input_shape[0] if "COUT" not in self.config else self.config["COUT"]
        return nn.Linear(cin, cout)

    def get_output_shape(self):
        cout = self.input_shape[0] if "COUT" not in self.config else self.config["COUT"]
        return [cout] + self.input_shape[1:]

#-------------------- activation function --------------------#

class Relu(BaseOperator):
    def get_model(self):
        return nn.ReLU()


class Relu6(BaseOperator):
    def get_model(self):
        return nn.ReLU()


class Sigmoid(BaseOperator):
    def get_model(self):
        return nn.Sigmoid()


class Hswish(BaseOperator):
    def get_model(self):
        return nn.Hardswish()

#---------------------- basic operation ----------------------#

class Reshape(BaseOperator):
    def get_model(self):
        if len(self.input_shape) == 3:
            self.output_shape = [self.input_shape[1], self.input_shape[2], self.input_shape[0]]
            def func(inputs):
                return torch.reshape(inputs, [1] + self.output_shape)
        else:
            self.output_shape = [1, 2, int(self.input_shape[0] / 2)]
            def func(inputs):
                return torch.reshape(inputs, [1] + self.output_shape)
        return func

    def get_output_shape(self):
        return self.output_shape


class Add(BaseOperator):
    def get_model(self):
        def func(inputs):
            return torch.add(inputs[0], inputs[1])
        return func

    def get_output_shape(self):
        if len(self.input_shape) == 2 and type(self.input_shape[0]) == list:
            output_shape = self.input_shape[0]
        else:
            output_shape = self.input_shape
        return output_shape

    def get_is_two_inputs(self):
        return True


class Concat(BaseOperator):
    def get_model(self):
        def func(inputs):
            return torch.cat(tuple(inputs), dim=1)
        return func

    def get_output_shape(self):
        if len(self.input_shape) > 1 and type(self.input_shape[0]) == list: # e.g. [[3, 28, 28], [5, 28, 28]] -> [8, 28, 28]
            output_shape = [sum([i[0] for i in self.input_shape])] + self.input_shape[0][1:]
        elif len(self.input_shape) == 3: # e.g. [4, 28, 28] -> [8, 28, 28]
            output_shape = [self.input_shape[0] * 2] + self.input_shape[1:]
        else: # e.g. [1024] -> [2048]
            output_shape = [self.input_shape[0] * 2]
        return output_shape

    def get_is_two_inputs(self):
        return True


class Flatten(BaseOperator):
    def get_model(self):
        return nn.Flatten()

    def get_output_shape(self):
        return [int(np.prod(self.input_shape))]


class Split(BaseOperator):
    def get_model(self):
        cin = self.input_shape[0]
        def func(inputs):
            return torch.split(inputs, [cin // 2, cin - cin // 2], dim=1)
        return func

    def get_output_shape(self):
        return [self.input_shape[0] // 2, self.input_shape[1], self.input_shape[2]]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import random
import torch
import jsonlines
from .bench_dataset import bench_dataset
from nn_meter.utils import get_user_data_folder
from nn_meter.utils.import_package import try_import_dgl


RAW_DATA_URL = "https://github.com/microsoft/nn-Meter/releases/download/v1.0-data/datasets.zip"
__user_dataset_folder__ = os.path.join(get_user_data_folder(), 'dataset')


hws = [
    "cortexA76cpu_tflite21",
    "adreno640gpu_tflite21",
    "adreno630gpu_tflite21",
    "myriadvpu_openvino2019r2",
]


class GNNDataset(torch.utils.data.Dataset):
    def __init__(self, train=True, device="cortexA76cpu_tflite21", split_ratio=0.8):
        """
        Dataloader of the Latency Dataset

        Parameters
        ----------
        data_dir : string
            Path to save the downloaded dataset
        train: bool
            Get the train dataset or the test dataset
        device: string
            The Device type of the corresponding latency
        shuffle: bool
            If shuffle the dataset at the begining of an epoch
        batch_size: int
            Batch size.
        split_ratio: float
            The ratio to split the train dataset and the test dataset.
        """
        err_str = "Not supported device type"
        assert device in hws, err_str
        self.device = device
        self.train = train
        self.split_ratio = split_ratio
        self.adjs = {}
        self.attrs = {}
        self.nodename2id = {}
        self.id2nodename = {}
        self.op_types = set()
        self.opname2id = {}
        self.raw_data = {}
        self.name_list = []
        self.latencies = {}
        self.data_dir = bench_dataset(data_folder=__user_dataset_folder__)
        self.load_model_archs_and_latencies(self.data_dir)
        self.construct_attrs()
        self.name_list = list(filter(lambda x: x in self.latencies, self.name_list))

    def load_model_archs_and_latencies(self, data_dir):
        for filename in data_dir:
            self.load_model(filename)

    def load_model(self, fpath):
        """
        Load a concrete model type.
        """
        assert os.path.exists(fpath), '{} does not exists'.format(fpath)

        with jsonlines.open(fpath) as reader:
            _names = []
            for obj in reader:
                if obj[self.device]:
                    _names.append(obj['id'])
                    self.latencies[obj['id']] = float(obj[self.device])

            _names = sorted(_names)
            split_ratio = self.split_ratio if self.train else 1-self.split_ratio
            count = int(len(_names) * split_ratio)

            if self.train:
                _model_names = _names[:count]
            else:
                _model_names = _names[-1*count:]

            self.name_list.extend(_model_names)

        with jsonlines.open(fpath) as reader:
            for obj in reader:
                if obj['id'] in _model_names:
                    model_name = obj['id']
                    model_data = obj['graph']
                    self.parse_model(model_name, model_data)
                    self.raw_data[model_name] = model_data
    
    def construct_attrs(self):
        """
        Construct the attributes matrix for each model.
        Attributes tensor:
        one-hot encoded type + input_channel , output_channel,
        input_h, input_w + kernel_size + stride
        """
        op_types_list = list(sorted(self.op_types))
        for i, _op in enumerate(op_types_list):
            self.opname2id[_op] = i
        n_op_type = len(self.op_types)
        attr_len = n_op_type + 6
        for model_name in self.raw_data:
            n_node = len(self.raw_data[model_name])
            # print("Model: ", model_name, " Number of Nodes: ", n_node)
            t_attr = torch.zeros(n_node, attr_len)
            for node in self.raw_data[model_name]:
                node_attr = self.raw_data[model_name][node]
                nid = self.nodename2id[model_name][node]
                op_type = node_attr['attr']['type']
                op_id = self.opname2id[op_type]
                t_attr[nid][op_id] = 1
                other_attrs = self.parse_node(model_name, node)
                t_attr[nid][-6:] = other_attrs
            self.attrs[model_name] = t_attr

    def parse_node(self, model_name, node_name):
        """
        Parse the attributes of specified node
        Get the input_c, output_c, input_h, input_w, kernel_size, stride
        of this node. Note: filled with 0 by default if this doesn't have
        coressponding attribute.
        """
        node_data = self.raw_data[model_name][node_name]
        t_attr = torch.zeros(6)
        op_type = node_data['attr']['type']
        if op_type =='Conv2D':
            weight_shape = node_data['attr']['attr']['weight_shape']
            kernel_size, _, in_c, out_c = weight_shape
            stride, _= node_data['attr']['attr']['strides']
            _, h, w, _ = node_data['attr']['output_shape'][0]
            t_attr = torch.tensor([in_c, out_c, h, w, kernel_size, stride])
        elif op_type == 'DepthwiseConv2dNative':
            weight_shape = node_data['attr']['attr']['weight_shape']
            kernel_size, _, in_c, out_c = weight_shape
            stride, _= node_data['attr']['attr']['strides']
            _, h, w, _ = node_data['attr']['output_shape'][0]
            t_attr = torch.tensor([in_c, out_c, h, w, kernel_size, stride])
        elif op_type == 'MatMul':
            in_node = node_data['inbounds'][0]
            in_shape = self.raw_data[model_name][in_node]['attr']['output_shape'][0]
            in_c = in_shape[-1]
            out_c = node_data['attr']['output_shape'][0][-1]
            t_attr[0] = in_c
            t_attr[1] = out_c
        elif len(node_data['inbounds']):
            in_node = node_data['inbounds'][0]
            h, w, in_c, out_c = 0, 0, 0, 0
            in_shape = self.raw_data[model_name][in_node]['attr']['output_shape'][0]
            in_c = in_shape[-1]
            if 'ConCat' in op_type:
                for i in range(1, len(node_data['in_bounds'])):
                    in_shape = self.raw_data[node_data['in_bounds']
                                             [i]]['attr']['output_shape'][0]
                    in_c += in_shape[-1]
            if len(node_data['attr']['output_shape']):
                out_shape = node_data['attr']['output_shape'][0]
                # N, H, W, C
                out_c = out_shape[-1]
                if len(out_shape) == 4:
                    h, w = out_shape[1], out_shape[2]
            t_attr[-6:-2] = torch.tensor([in_c, out_c, h, w])

        return t_attr

    def parse_model(self, model_name, model_data):
        """
        Parse the model data and build the adjacent matrixes
        """
        n_nodes = len(model_data)
        m_adj = torch.zeros(n_nodes, n_nodes, dtype=torch.int32)
        id2name = {}
        name2id = {}
        tmp_node_id = 0
        # build the mapping between the node name and node id

        for node_name in model_data.keys():
            id2name[tmp_node_id] = node_name
            name2id[node_name] = tmp_node_id
            op_type = model_data[node_name]['attr']['type']
            self.op_types.add(op_type)
            tmp_node_id += 1

        for node_name in model_data:
            cur_id = name2id[node_name]
            for node in model_data[node_name]['inbounds']:
                if node not in name2id:
                    # weight node
                    continue
                in_id = name2id[node]
                m_adj[in_id][cur_id] = 1
            for node in model_data[node_name]['outbounds']:
                if node not in name2id:
                    # weight node
                    continue
                out_id = name2id[node]
                m_adj[cur_id][out_id] = 1
        
        for idx in range(n_nodes):
            m_adj[idx][idx] = 1

        self.adjs[model_name] = m_adj
        self.nodename2id[model_name] = name2id
        self.id2nodename[model_name] = id2name

    def __getitem__(self, index):
        model_name = self.name_list[index]
        return (self.adjs[model_name], self.attrs[model_name]), self.latencies[model_name], self.op_types

    def __len__(self):
        return len(self.name_list)


class GNNDataloader(torch.utils.data.DataLoader):
    def __init__(self, dataset, shuffle=False, batchsize=1):
        self.dataset = dataset
        self.op_num = len(dataset.op_types)
        self.shuffle = shuffle
        self.batchsize = batchsize
        self.length = len(self.dataset)
        self.indexes = list(range(self.length))
        self.pos = 0
        self.graphs = {}
        self.latencies = {}
        self.construct_graphs()

    def construct_graphs(self):
        dgl = try_import_dgl()
        for gid in range(self.length):
            (adj, attrs), latency, op_types = self.dataset[gid]
            u, v = torch.nonzero(adj, as_tuple=True)
            graph = dgl.graph((u, v))
            MAX_NORM = torch.tensor([1]*len(op_types) + [6963, 6963, 224, 224, 11, 4])
            attrs = attrs / MAX_NORM
            graph.ndata['h'] = attrs
            self.graphs[gid] = graph
            self.latencies[gid] = latency

    def __iter__(self):
        if self.shuffle:
            random.shuffle(self.indexes)
        self.pos = 0
        return self

    def __len__(self):
        return self.length

    def __next__(self):
        dgl = try_import_dgl()
        start = self.pos
        end = min(start + self.batchsize, self.length)
        self.pos = end
        if end - start <= 0:
            raise StopIteration
        batch_indexes = self.indexes[start:end]
        batch_graphs = [self.graphs[i] for i in batch_indexes]
        batch_latencies = [self.latencies[i] for i in batch_indexes]
        return torch.tensor(batch_latencies), dgl.batch(batch_graphs)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import re
import sys

import tensorflow as tf
from google.protobuf import text_format
from tensorflow import io


def dec2octpkg4(input):
    octPkgStr = ''
    for i in range(4):
        octPkgStr = octPkgStr + oct((input >> (i * 8)) %
                                    256).replace('0o', '\\')
    return octPkgStr


KEEP_DIM_PATCH =\
    '''
  attr {
    key: "keep_dims"
    value {
      b: {KEEP_DIM}
    }
  }
'''

REDUCE_DIM_PATCH =\
    '''
node {
  name: "reshape/Reshape/shape"
  op: "Const"
  attr {
    key: "dtype"
    value {
      type: DT_INT32
    }
  }
  attr {
    key: "value"
    value {
      tensor {
        dtype: DT_INT32
        tensor_shape {
          dim {
            size: 2
          }
        }
        tensor_content: "{SHAPE}"
      }
    }
  }
}

node {
  name: "reshape/Reshape"
  op: "Reshape"
  input: "{INPUT_TENSOR_NAME}"
  input: "reshape/Reshape/shape"
  attr {
    key: "T"
    value {
      type: DT_FLOAT
    }
  }
  attr {
    key: "Tshape"
    value {
      type: DT_INT32
    }
  }
}
'''

SWISH_PATCH =\
    '''
node {
  name: "{NAME}/Sigmoid"
  op: "Sigmoid"
  input: "{INPUT}"
  attr {
    key: "T"
    value {
      type: DT_FLOAT
    }
  }
}

node {
  name: "{NAME}/mul"
  op: "Mul"
  input: "{INPUT}"
  input: "{NAME}/Sigmoid"
  attr {
    key: "T"
    value {
      type: DT_FLOAT
    }
  }
}
'''

MEAN_PATCH =\
    '''
node {
  name: "{NAME}"
  op: "AvgPool"
  input: "{INPUT}"
  attr {
    key: "T"
    value {
      type: DT_FLOAT
    }
  }
  attr {
    key: "data_format"
    value {
      s: "NHWC"
    }
  }
  attr {
    key: "ksize"
    value {
      list {
        i: 1
        i: {KERNEL_SIZE}
        i: {KERNEL_SIZE}
        i: 1
      }
    }
  }
  attr {
    key: "padding"
    value {
      s: "VALID"
    }
  }
  attr {
    key: "strides"
    value {
      list {
        i: 1
        i: 1
        i: 1
        i: 1
      }
    }
  }
}
'''

EXPLICT_PAD_ATTR_REGEX = r'attr {\n[\s]+key: "explicit_paddings"\n[\s]'\
    r'+value {\n[\s]+list {\n[\s]+}\n[\s]+}\n[\s]+}'

U_KEY_ATTR_REGEX = r'attr {\n[\s]+key: "U"\n[\s]+value {\n[\s]+type: DT_FLOAT\n[\s]+}\n[\s]+}'

OUTPUT_SHAPE_REGEX_3 = r'([\s]+attr {\n[\s]+key: "_output_shapes"\n[\s]+value {\n[\s]+list {\n['\
    r'\s]+)(shape[\s]+{[\s]+([\s]+(dim[\s]+{\s+size:[\s]+[0-9]+[\s]+})|([\s]+'\
    r'unknown_rank: \w+([\s]+})+))+([\s]+}[\s]+)+)+([\s]})+'

OUTPUT_SHAPE_REGEX_1 = r'attr {\n[\s]+key: "_output_shapes"\n[\s]+value {\n[\s]+list {\n[\s]+shape '\
    r'{\n[\s]+dim {\n[\s]+size: [0-9]+\n[\s]+}\n[\s]+dim {\n[\s]+size: [0-9]+\n[\s]+}'\
    r'\n[\s]+dim {\n[\s]+size: [0-9]+\n[\s]+}\n[\s]+dim {\n[\s]+size: [0-9]+\n[\s]+}\n'\
    r'[\s]+}\n[\s]+shape {\n[\s]+dim {\n[\s]+size: [0-9]+\n[\s]+}\n[\s]+}\n[\s]+shape'\
    r' {\n[\s]+dim {\n[\s]+size: [0-9]+\n[\s]+}\n[\s]+}\n[\s]+shape {\n[\s]+dim {\n[\s'\
    r']+size: [0-9]+\n[\s]+}\n[\s]+}\n[\s]+shape {\n[\s]+dim {\n[\s]+size: [0-9]+\n[\s]'\
    r'+}\n[\s]+}\n[\s]+shape {\n[\s]+unknown_rank: true\n[\s]+}\n[\s]+}\n[\s]+}\n[\s]+}'

OUTPUT_SHAPE_REGEX_2 = r'[\s]+attr {\n[\s]+key: "_output_shapes"\n[\s]+value {\n[\s]+list {\n[\s]+shape '\
    r'{([\s]+dim[\s]+{\n[\s]+size:[\s]+[0-9]+\n[\s]+}\n)+([\s]+}\n)+'

REDUCTION_IDENCE_REGEX = r'node[\s]+{\n[\s]+name:[\s]+\"[^"]+reduction_indices'\
    r'\"\n[\s]+op:[\s]+\"Const\"[\s]+[\s]+attr[\s]+{\n[\s]+k'\
    r'ey:[\s]+\"dtype\"\n[\s]+value[\s]+{\n[\s]+type:[\s]+DT'\
    r'_INT32\n[\s]+}\n[\s]+}\n[\s]+attr[\s]+{\n[\s]+key:[\s]'\
    r'+\"value\"\n[\s]+value[\s]+{\n[\s]+tensor[\s]+{\n[\s]+'\
    r'dtype:[\s]+DT_INT32\n[\s]+tensor_shape[\s]+{\n[\s]+dim'\
    r'[\s]+{\n[\s]+size:[\s]+2\n[\s]+}\n[\s]+}\n[\s]+tensor_'\
    r'content:[\s]+\"\\001\\000\\000\\000\\002\\000\\000\\00'\
    r'0\"\n+([\s]+}\n[\s]+)(}\n[\s]+}\n})'


def pbtxt_processing(content):
    if content.find('explicit_paddings') != -1:
        print('Find unsupported attr: explicit_paddings, removing...\n')
        content = re.sub(EXPLICT_PAD_ATTR_REGEX, '', content)

    meanCounter = content.count('op: "Mean"')
    if meanCounter > 1:  # May not work
        endMeanNodeName = ""
        print(f'Find semi-supported op: Mean presented {meanCounter} times in pb, patching...')
        while True:
            meanOpLoc = content.find('op: "Mean"')
            if meanOpLoc == -1:
                break
            nodeNameLoc = content.rfind('name', 0, meanOpLoc)
            nodeNameDLoc = content.find('"', nodeNameLoc)
            nodeName = content[nodeNameDLoc +
                               1:content.find('"', nodeNameDLoc + 1)]

            nodeInputLoc = content.find('input', meanOpLoc)
            nodeInputDLoc = content.find('"', nodeInputLoc)
            nodeInputName = content[nodeInputDLoc +
                                    1:content.find('"', nodeInputDLoc + 1)]

            inputNodeNameLoc = content.find(f'name: "{nodeInputName}"')
            inputNodeEnd = content.find('node', inputNodeNameLoc)
            inputNodeShape = re.findall(
                r'[\d]+\n', content[inputNodeNameLoc:inputNodeEnd])

            if len(inputNodeShape) != 4:
                print(
                    f'Unexpected happened in shape inference, infered shape: {inputNodeShape} in node [{nodeName}]')
                sys.exit(-1)

            for i in range(len(inputNodeShape)):
                inputNodeShape[i] = int(inputNodeShape[i].replace('\n', ''))
            print(
                f'Found Node name: {nodeName}, Input Shape: {inputNodeShape}\nPatching the Mean operator...')

            nodeStart = content.rfind('{', 0, nodeNameLoc)
            nodeEnd = content.find('node', nodeNameLoc)

            if content[nodeStart:nodeEnd].find(
                    KEEP_DIM_PATCH.replace('{KEEP_DIM}', 'false')) != -1:
                print('Find reduce mean at top, ignore and break.')
                endMeanNodeName = nodeName
                break

            print(f'Generating the patcher, node input: {nodeInputName}')
            patcher = MEAN_PATCH.replace('{NAME}', nodeName)
            patcher = patcher.replace('{INPUT}', nodeInputName)
            patcher = patcher.replace('{KERNEL_SIZE}', str(inputNodeShape[1]))

            print('Inserting patch and removing the Mean node...\n')
            content = content[:content.rfind(
                'node', 0, nodeStart)] + patcher + content[nodeEnd:]

        print('Removing unused const.\n')
        content = re.sub(REDUCTION_IDENCE_REGEX, '', content)

        while True:
            indecOpLoc = content.find('reduction_indices')
            if indecOpLoc == -1:
                break
            indecNameLoc = content.rfind('name', 0, indecOpLoc)
            indecStart = content.rfind('{', 0, indecNameLoc)
            indecEnd = content.find('node', indecNameLoc)
            if content[indecStart:indecEnd].find(endMeanNodeName) != -1:
                break
            content = content[:content.rfind(
                'node', 0, indecStart)] + content[indecEnd:]

    if content.find('AddV2') != -1:
        print('Find unsupported op: AddV2, patching...\n')
        content = content.replace('AddV2', 'Add')

    if content.find(KEEP_DIM_PATCH.replace('{KEEP_DIM}', 'false')) != -1:
        print('Find unsupported op: reduce_dim=false, patching...')

        while True:
            keepDimLoc = content.find(
                KEEP_DIM_PATCH.replace(
                    '{KEEP_DIM}', 'false'))
            if keepDimLoc == -1:
                break

            nodeNameLoc = content.rfind('name', 0, keepDimLoc)
            nodeNameDLoc = content.find('"', nodeNameLoc)
            nodeName = content[nodeNameDLoc +
                               1:content.find('"', nodeNameDLoc + 1)]
            print(
                f'Found Node name: {nodeName}, Output Shape: {OUTPUT_FILTER}, Oct: {dec2octpkg4(OUTPUT_FILTER)}')
            print('Patching the Mean operator...')

            nodeEnd = content.find('node', nodeNameLoc)
            content = content.replace(
                f'input: "{nodeName}"',
                'input: "reshape/Reshape"')

            patcher = REDUCE_DIM_PATCH.replace('{INPUT_TENSOR_NAME}', nodeName)
            patcher = patcher.replace(
                '{SHAPE}', f'\\377\\377\\377\\377{dec2octpkg4(OUTPUT_FILTER)}')

            content = content[:nodeEnd] + patcher + content[nodeEnd:]

            content = content.replace(
                KEEP_DIM_PATCH.replace(
                    '{KEEP_DIM}', 'false'), KEEP_DIM_PATCH.replace(
                    '{KEEP_DIM}', 'true'))
            print('Modified reduce_dim=true...\n')

    if content.find('FusedBatchNormV3') != -1:
        print('Find unsupported op: FusedBatchNormV3, patching...\n')
        content = content.replace('FusedBatchNormV3', 'FusedBatchNorm')
        content = re.sub(U_KEY_ATTR_REGEX, '', content)
        content = re.sub(OUTPUT_SHAPE_REGEX_1, '', content)
        content = re.sub(OUTPUT_SHAPE_REGEX_2, '', content)

    if content.find('op: "swish_f32"') != -1:
        print('Find unsupported op: swish_f32, patching...')
        while True:
            swishOpLoc = content.find('op: "swish_f32"')
            if swishOpLoc == -1:
                break
            nodeNameLoc = content.rfind('name', 0, swishOpLoc)
            nodeNameDLoc = content.find('"', nodeNameLoc)
            nodeName = content[nodeNameDLoc +
                               1:content.find('"', nodeNameDLoc + 1)]

            nodeInputLoc = content.find('input', swishOpLoc)
            nodeInputDLoc = content.find('"', nodeInputLoc)
            nodeInputName = content[nodeInputDLoc +
                                    1:content.find('"', nodeInputDLoc + 1)]

            print(
                f'Found Node name: {nodeName}\nPatching the swish_f32 operator...')

            nodeStart = content.rfind('{', 0, nodeNameLoc)
            nodeEnd = content.find('node', nodeNameLoc)

            print(f'Generating the patcher, node input: {nodeInputName}')
            patcher = SWISH_PATCH.replace('{NAME}', nodeName)
            patcher = patcher.replace('{INPUT}', nodeInputName)

            print('Inserting patch and removing the swish_f32 node...')
            content = content[:content.rfind(
                'node', 0, nodeStart)] + patcher + content[nodeEnd:]

            print('Reconnecting the graph...\n')
            content = content.replace(
                f'input: "{nodeName}"',
                f'input: "{nodeName}/mul"')

    return content


FILE_NAME = sys.argv[1]
PBTXT_FILE_NAME = FILE_NAME.replace('.pb', '.pbtxt')

OUTPUT_FILTER = 1280
if len(sys.argv) > 2:
    OUTPUT_FILTER = int(sys.argv[2])

if not os.path.isfile(PBTXT_FILE_NAME):
    f = open(FILE_NAME, 'rb')
    GRAPH_DEF = tf.compat.v1.get_default_graph().as_graph_def(add_shapes=True)
    GRAPH_DEF.ParseFromString(f.read())
    f.close()

    tf.import_graph_def(GRAPH_DEF, name='')
    io.write_graph(GRAPH_DEF, '', PBTXT_FILE_NAME, as_text=True)
else:

    GRAPH_DEF = tf.get_default_graph().as_graph_def(add_shapes=True)

    FILE_CONTENT = pbtxt_processing(open(PBTXT_FILE_NAME, 'r').read())

    print('Content check OK, start merging...')

    text_format.Merge(FILE_CONTENT, GRAPH_DEF)
    io.write_graph(GRAPH_DEF,
                   os.path.dirname(FILE_NAME),
                   os.path.basename(FILE_NAME).split('.')[0] + '_patched.pb',
                   as_text=False)
    
    os.remove(PBTXT_FILE_NAME)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import json
import time
import logging
from . import builder_config
from .utils import save_profiled_results, merge_info
from nn_meter.builder.backends import connect_backend
logging = logging.getLogger("nn-Meter")


def convert_models(backend, models, mode = 'predbuild', broken_point_mode = False):
    """ convert the model to the needed format by backend, in order to increase efficiency when profiling on device.

    @params:

    backend (subclass instance of BaseBackend): applied backend instance

    models (str or dict): the Dict of models or the path of the json file about models information 

    mode (str): the mode for running models, including ['ruletest', 'predbuild']

    broken_point_mode (boolean): broken_point_mode will skip all models have attributes "converted_model"

    """
    if isinstance(models, str):
        save_name = os.path.basename(models)
        with open(models, 'r') as fp:
            models = json.load(fp)
    else:
        save_name = "converted_results.json"

    workspace_path = builder_config.get('WORKSPACE', mode)
    model_save_path = os.path.join(workspace_path, 'testcases' if mode == 'ruletest' else 'kernels')
    os.makedirs(model_save_path, exist_ok=True)
    res_save_path = os.path.join(workspace_path, "results")
    os.makedirs(res_save_path, exist_ok=True)

    # convert models
    count = 0
    info_save_path = os.path.join(res_save_path, save_name)
    error_save_path = os.path.join(res_save_path, "convert_error.log")
    for module in models.values():
        for id, model in module.items():
            if broken_point_mode and 'converted_model' in model:
                continue
            try:
                model_path = model['model']
                converted_model = backend.convert_model(model_path, model_save_path, model['shapes'])
                model['converted_model'] = converted_model
                count += 1
            except Exception as e:
                open(error_save_path, 'a').write(f"{id}: {e}\n")

            # save information to json file for per 50 models
            if count % 50 == 0:
                with open(info_save_path, 'w') as fp:
                    json.dump(models, fp, indent=4)
                logging.keyinfo(f"{count} models complete. Still converting... Save the intermediate results to {info_save_path} ")

    # save information to json file
    with open(info_save_path, 'w') as fp:
        json.dump(models, fp, indent=4)
    logging.keyinfo(f"Complete converting all {count} models. Save the results to {info_save_path} " \
                    f"Failed information are saved in {error_save_path} (if any)")

    return models


def profile_models(backend, models, mode = 'ruletest', metrics = ["latency"], save_name = "profiled_results.json",
                   have_converted = False, log_frequency = 50, broken_point_mode = False, **kwargs):
    """ run models with given backend and return latency of testcase models

    @params:

    backend (subclass instance of BaseBackend): applied backend instance

    models (str or dict): the Dict of models or the path of the json file about models information 

    mode (str): the mode for running models, including ['ruletest', 'predbuild']

    metrics (list): required metrics to report. We only support latency for metric by now.

    save_name (str): the save name to store profiled results. The whole path should be `<workspace>/<mode-folder>/results/<save-name>`

    have_converted (boolean): if the model have been converted to the needed format by backend, the model will not be converted
        before profiling. The model path of `model['converted_model']` will be profiled on device directly. The conversion of
        model could be done by appling `nn_meter.builder.convert_models`

    broken_point_mode (boolean): broken_point_mode will check file in `<workspace>/<mode-folder>/results/<save-name>` (if the file exists)
        and skip all models already have attributes "latency"

    **kwargs: arguments for profiler, such as `taskset` and `close_xnnpack` in TFLite profiler
    """
    if isinstance(models, str):
        with open(models, 'r') as fp:
            models = json.load(fp)

    workspace_path = builder_config.get('WORKSPACE', mode)
    model_save_path = os.path.join(workspace_path, 'testcases' if mode == 'ruletest' else 'kernels')
    os.makedirs(model_save_path, exist_ok=True)
    res_save_path = os.path.join(workspace_path, "results")
    os.makedirs(res_save_path, exist_ok=True)
    info_save_path = os.path.join(res_save_path, save_name)

    # in broken point model, if the output file `<workspace>/<mode-folder>/results/<save-name>` exists,
    # load the existing latency and skip these model in profiling
    if broken_point_mode and os.path.isfile(info_save_path):
        from nn_meter.builder.backend_meta.utils import read_profiled_results
        with open(info_save_path, 'r') as fp:
            profiled_models = read_profiled_results(json.load(fp))
        for module_key, module in models.items():
            if module_key not in profiled_models:
                continue
            for id, model in module.items():
                if id in profiled_models[module_key]:
                    model.update(profiled_models[module_key][id])

    # profile models and get metric results
    count = 0    
    error_save_path = os.path.join(res_save_path, "profile_error.log")
    detail = builder_config.get('DETAIL', mode)
    save_name = save_name or "profiled_results.json"
    logging.info("Profiling ...")
    for module in models.values():
        for id, model in module.items():
            if broken_point_mode and 'latency' in model and model['latency'].avg != 0:
                continue
            if have_converted: # the models have been converted for the backend
                try:
                    model_path = model['converted_model']
                    profiled_res = backend.profile(model_path, metrics, model['shapes'], **kwargs)
                    for metric in metrics:
                        model[metric] = profiled_res[metric]
                    time.sleep(0.2)
                    count += 1
                except Exception as e:
                    open(error_save_path, 'a').write(f"{id}: {e}\n")
            else: # the models have not been converted
                try:
                    model_path = model['model']
                    profiled_res = backend.profile_model_file(model_path, model_save_path, model['shapes'], metrics, **kwargs)
                    for metric in metrics:
                        model[metric] = profiled_res[metric]
                    time.sleep(0.2)
                    count += 1
                except Exception as e:
                    open(error_save_path, 'a').write(f"{id}: {e}\n")

            # save information to json file for per 50 models
            if count > 0 and count % log_frequency == 0:
                save_profiled_results(models, info_save_path, detail, metrics)
                logging.keyinfo(f"{count} models complete. Still profiling... Save the intermediate results to {info_save_path} ")

    # save information to json file
    save_profiled_results(models, info_save_path, detail, metrics)    
    logging.keyinfo(f"All {count} models profiling complete. Save all success profiled results to {info_save_path} " \
                    f"Failed information are saved in {error_save_path} (if any)")

    return models


def sample_and_profile_kernel_data(kernel_type, sample_num, backend, sampling_mode = 'prior', configs = None, mark = '', detail = True,
                                   metrics = ["latency"], **kwargs):
    ''' sample kernel configs and profile kernel model based on configs
    '''
    from nn_meter.builder.kernel_predictor_builder import generate_config_sample

    # sample configs for kernel and generate models
    models = generate_config_sample(kernel_type, sample_num, mark=mark, 
                                     sampling_mode=sampling_mode, configs=configs)

    # connect to backend, run models and get latency
    backend = connect_backend(backend_name=backend)
    profiled_results = profile_models(backend, models, mode='predbuild', metrics=metrics, save_name=f"profiled_{kernel_type}.json")
    return profiled_results


def build_predictor_for_kernel(kernel_type, backend, init_sample_num = 1000, finegrained_sample_num = 10,
                               iteration = 5, error_threshold = 0.1, predict_label = "latency", mark = ""):
    """ 
    Build latency predictor for given kernel. This method contains three main steps:
    1. sample kernel configs and profile kernel model based on configs;
    2. initialize latency predictor of kernel based on the profiled data;
    3. adopt adaptive sampler with iteratively doing step 1 for finegrained sampling to improve predictor performance

    @params
    
    kernel_type (str): the type of kernel
    
    backend (str): the name of backend instance to profile models
    
    init_sample_num (int, optional): the data size for predictor initialization. Defaults to 1000.
    
    finegrained_sample_num (int, optional): the data size for adaptive sampling. For each data with error higher than 
        error_threshold, number of `finegrained_sample_num` data will be generated based the the large error data. Defaults to 10.

    iteration (int, optional): the iteration for sampling and training. Predictor training based on initial sampling is regarded as
    iteration 1, thus `iteration == 2` means one iteration for adaptive sampling. Defaults to 5.

    error_threshold (float, optional): the threshold of large error. Defaults to 0.1.

    predict_label (str): the predicting label to build kernel predictor. Defaults to "latency"
 
    """
    from nn_meter.builder.kernel_predictor_builder import build_predictor_by_data
    workspace_path = builder_config.get('WORKSPACE', 'predbuild')
    mark = mark if mark == "" else "_" + mark

    # init predictor builder with prior data sampler
    kernel_data = sample_and_profile_kernel_data(kernel_type, init_sample_num, backend, sampling_mode='prior', mark=f'prior{mark}')

    # use current sampled data to build regression model, and locate data with large errors in testset
    predictor, acc10, error_configs = build_predictor_by_data(kernel_type, kernel_data, backend, error_threshold=error_threshold, mark=f'prior{mark}',
                                                              save_path=os.path.join(workspace_path, "results"), predict_label=predict_label)
    logging.keyinfo(f'Iteration 0: acc10 {acc10}, error_configs number: {len(error_configs)}')

    for i in range(1, iteration):
        # finegrained sampling and profiling for large error data
        new_kernel_data = sample_and_profile_kernel_data(kernel_type, finegrained_sample_num, backend,
                                                         sampling_mode='finegrained', configs=error_configs, mark=f'finegrained{i}{mark}')

        # merge finegrained data with previous data and build new regression model
        kernel_data = merge_info(new_info=new_kernel_data, prev_info=kernel_data)
        predictor, acc10, error_configs = build_predictor_by_data(kernel_type, kernel_data, backend, error_threshold=error_threshold, mark=f'finegrained{i}{mark}',
                                                                  save_path=os.path.join(workspace_path, "results"), predict_label=predict_label)
        logging.keyinfo(f'Iteration {i}: acc10 {acc10}, error_configs number: {len(error_configs)}')

    return predictor, kernel_data


def build_initial_predictor_by_data(kernel_type, backend = None, init_sample_num = 20, error_threshold = 0.1, mark = '', predict_label = "latency"):
    return build_predictor_for_kernel(kernel_type, backend, init_sample_num=init_sample_num, iteration=1, error_threshold=error_threshold, predict_label=predict_label, mark=mark)


def build_adaptive_predictor_by_data(kernel_type, kernel_data, backend = None, finegrained_sample_num = 20, error_threshold = 0.1, mark = '', predict_label = "latency"):
    """ Run adaptive sampler in one iteration based 
    """
    workspace_path = builder_config.get('WORKSPACE', 'predbuild')
    save_path = os.path.join(workspace_path, "results")
    mark = 'finegrained' if mark == "" else f"finegrained_{mark}"

    from nn_meter.builder.kernel_predictor_builder import build_predictor_by_data, collect_kernel_data
    _, _, error_configs = build_predictor_by_data(kernel_type, kernel_data, backend = backend, error_threshold=error_threshold, save_path=None, predict_label=predict_label)
    new_kernel_data = sample_and_profile_kernel_data(kernel_type, finegrained_sample_num, backend,
                                                     sampling_mode='finegrained', configs=error_configs, mark=mark)

    # merge finegrained data with previous data and build new regression model
    kernel_data = merge_info(new_info=new_kernel_data, prev_info=collect_kernel_data(kernel_data))
    predictor, acc10, error_configs = build_predictor_by_data(kernel_type, kernel_data, backend, error_threshold=error_threshold,
                                                              mark=mark, save_path=save_path, predict_label=predict_label)
    logging.keyinfo(f'{mark}: acc10 {acc10}, error_configs number: {len(error_configs)}')
    return predictor, kernel_data


def build_latency_predictor(backend):
    """ 
    Build latency predictor for all kernel in `<workspace-path>/configs/predictorbuild_config.yaml`

    @params

    backend (str): the name of backend instance to profile models

    """
    kernels = builder_config.get("KERNELS", 'predbuild')

    for kernel_type in kernels:
        init_sample_num = kernels[kernel_type]["INIT_SAMPLE_NUM"]
        finegrained_sample_num = kernels[kernel_type]["FINEGRAINED_SAMPLE_NUM"]
        iteration = kernels[kernel_type]["ITERATION"]
        error_threshold = kernels[kernel_type]["ERROR_THRESHOLD"]
        build_predictor_for_kernel(
            kernel_type, backend, 
            init_sample_num = init_sample_num,
            finegrained_sample_num = finegrained_sample_num,
            iteration = iteration,
            error_threshold = error_threshold
            )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import math
import copy
from typing import List


class ProfiledResults:
    def __init__(self, results=None):
        """
        Initialize the profiled results of test models running on backends.
        
        @params:
        
        results: Dict
            The profiled results, with Dict key to be the metric name. Metrics 
            include: latency, peak/average power, energy, memory, etc.
        """
        self.data = {}
        for metric, value in results.items():
            self.data[metric] = value
    
    def set(self, metric, value):
        """ Set metric value by its name
        """
        self.data[metric] = value

    def get(self, metrics):
        ''' Get metric value by calling the name of the metric.
        '''
        if not isinstance(metrics, List):
            metrics = [metrics]
        result = {}
        for metric in metrics:
            if metric in self.data:
                result[metric] = self.data[metric]
            else:
                raise AttributeError(f"Unsupported metric {metric}.")
        return result
    
    def _dump(self):
        return {metric: str(value) for metric, value in self.data}


class Latency:
    def __init__(self, avg=0, std=0):
        if isinstance(avg, str):
            avg, std = avg.split('+-')
            self.avg = float(avg)
            self.std = float(std)
        elif isinstance(avg, Latency):
            self.avg, self.std = avg.avg, avg.std
        else:
            self.avg = avg
            self.std = std

    def __str__(self):
        return f'{self.avg} +- {self.std}'

    def __add__(self, rhs):
        if isinstance(rhs, Latency):
            return Latency(self.avg + rhs.avg, math.sqrt(self.std ** 2 + rhs.std ** 2))
        else:
            return Latency(self.avg + rhs, self.std)
    
    def __radd__(self, lhs):
        return self.__add__(lhs)

    def __mul__(self, rhs):
        return Latency(self.avg * rhs, self.std * rhs)

    def __rmul__(self, lhs):
        return self.__mul__(lhs)

    def __le__(self, rhs):
        return self.avg < rhs.avg
    
    def __gt__(self, rhs):
        return self.avg > rhs.avg

    def __neg__(self):
        return Latency(-self.avg, -self.std)

    def __sub__(self, rhs):
        return self + rhs.__neg__()


def dump_profiled_results(results, detail = False, metrics = ["latency"]):
    ''' convert Latency instance to string and return profiled results

    @params

    detail: if False, only metrics result will be dumped to the profiled results. Otherwise models information
        will be dumpled, too.
    '''
    dumped_results = {}
    for module_key, module in results.items():
        dumped_results[module_key] = {}
        for model_key, model in module.items():
            dumped_results[module_key][model_key] = {}
            if detail:
                for info_key, info in model.items():
                    if info_key == 'latency':
                        dumped_results[module_key][model_key]['latency'] = str(model['latency'])
                    else:
                        dumped_results[module_key][model_key][info_key] = info
            else:
                for info_key, info in model.items():
                    if info_key in metrics:
                        if info_key == 'latency':
                            dumped_results[module_key][model_key]['latency'] = str(model['latency'])
                        else:
                            dumped_results[module_key][model_key][info_key] = model[info_key]
    return dumped_results


def read_profiled_results(results):
    results_copy = copy.deepcopy(results)
    for item in results_copy.values():
        for model in item.values():
            if 'latency' in model:
                model['latency'] = Latency(model['latency'])
    return results_copy



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .build_predictor import build_predictor_by_data
from .extract_feature import get_data_by_profiled_results, BaseFeatureParser
from .utils import collect_kernel_data


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import json
import random
import string
import logging
from nn_meter.builder import builder_config
from nn_meter.builder.utils import merge_info
from .utils import get_sampler_for_kernel, generate_model_for_kernel
logging = logging.getLogger("nn-Meter")


class KernelGenerator:
    def __init__(self, kernel_type, sample_num, mark = ""):
        self.kernel_type = kernel_type
        self.sample_num = sample_num
        self.workspace_path = builder_config.get('WORKSPACE', 'predbuild')
        self.case_save_path = os.path.join(self.workspace_path, 'kernels')
        self.kernel_info = {kernel_type: {}}
        self.kernels = self.kernel_info[self.kernel_type]
        self.implement = builder_config.get('IMPLEMENT', 'predbuild')
        self.batch_size = builder_config.get('BATCH_SIZE', 'predbuild')
        self.model_suffix = "" if self.implement == 'tensorflow' else ".onnx"
        self.mark = mark
        os.makedirs(self.case_save_path, exist_ok=True)

    def generate_config(self, sampling_mode = 'prior', configs = None):
        sampled_cfgs = get_sampler_for_kernel(self.kernel_type, self.sample_num, sampling_mode, configs)
        for i in range(len(sampled_cfgs)):
            random_id = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(6))
            self.kernels[random_id] = {}
            self.kernels[random_id]['config'] = sampled_cfgs[i]

    def generate_kernel_by_cfg(self):
        """ generate tensorflow models for sampled data
        """
        kernel_type = self.kernel_type
        logging.info(f"building kernel for {kernel_type}...")
        count = 0
        error_save_path = os.path.join(self.workspace_path, 'results', 'generate_error.log')
        for id, value in self.kernels.items():
            model_path = os.path.join(self.case_save_path, ("_".join([kernel_type, self.mark, id]) + self.model_suffix))
            kernel_cfg = value['config']
            try:
                _, input_tensor_shape, config = generate_model_for_kernel(
                    kernel_type, kernel_cfg, save_path=model_path,
                    implement=self.implement, batch_size=self.batch_size
                )
                self.kernels[id] = {
                    'model': model_path,
                    'shapes': input_tensor_shape,
                    'config': config
                }
                count += 1
            except Exception as e:
                open(os.path.join(self.workspace_path, "results", "generate_error.log"), 'a').write(f"{id}: {e}\n")

        # save information to json file in incrementally mode
        info_save_path = os.path.join(self.workspace_path, "results", f"{kernel_type}_{self.mark}.json")
        new_kernels_info = merge_info(new_info=self.kernel_info, info_save_path=info_save_path)
        os.makedirs(os.path.dirname(info_save_path), exist_ok=True)
        with open(info_save_path, 'w') as fp:
            json.dump(new_kernels_info, fp, indent=4)
        logging.keyinfo(f"Generate {len(self.kernels)} kernels and save info to {info_save_path} " \
                        f"Failed information are saved in {error_save_path} (if any).")

    def run(self, sampling_mode = 'prior', configs = None):
        """ sample N configurations for target kernel, generate tensorflow keras model files.

        @params
        sampling_mode: path of the directory containing all experiment runs. choose from ['prior', 'finegrained']
        configs: init configs for finegrained sampling
        """
        # sample configs
        self.generate_config(sampling_mode, configs)
        
        # for all sampled configurations, save kernels info and generate tensorflow model files 
        self.generate_kernel_by_cfg()
        logging.info(f'Generate {len(self.kernels)} kernels with kernels model saved in {self.case_save_path}.')
        return self.kernel_info


def generate_config_sample(kernel_type, sample_num, mark = '', sampling_mode = 'prior', configs = None):
    """ Generate config sample and return sampled configs.

    @params
    kernel_type (str): type of kernel

    sample_num (int): the sampling number of configs

    mark (str, optional): the mark for the running results. Defaults to ''.

    sampling_mode (str, optional): the sampling mode for config generation, supporting mode includes 'prior' and 'finegrained'.
        Defaults to be 'prior'.

    configs (list, optional): is required when the sampling_mode=='finegrained'. The fingrained samples will based on the config 
        in `configs`. Defaults to None.

    """
    generator = KernelGenerator(kernel_type, sample_num, mark=mark)
    kernels_info = generator.run(sampling_mode=sampling_mode, configs=configs)

    return kernels_info



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import json


def merge_info(new_info, info_save_path = None, prev_info = None):
    ''' merge `new_info` with previous info and return the updated info. This method is used in two cases: 

    1. before save `new_info` to `info_save_path`, we need to check if the `info_save_path` is an existing file. If `info_save_path`
    exists, this method will help merge the previous info saved in `info_save_path` for a incrementally storage and avoid information
    loss. In this case, params `new_info` and `info_save_path` are needed.

    2. extend the dictionary of `prev_info` with `new_info`. In this case, params `new_info` and `prev_info` are needed.
    
    @params
    
    new_info (dict): new information
    
    info_save_path (str): the path to save the new info. We need to check if the path is empty and mantain the previous info
        in `info_save_path`.
    
    prev_info (dict): the previous information
    '''
    if (info_save_path == None and prev_info == None) or (info_save_path != None and prev_info != None):
        raise ValueError("One and only one params of `info_save_path` and `prev_info` is needed.")

    if info_save_path != None and os.path.isfile(info_save_path):
        with open(info_save_path, 'r') as fp:
            prev_info = json.load(fp)

    if prev_info == None:
        return new_info

    if isinstance(prev_info, str):
        with open(prev_info, 'r') as fp:
            prev_info = json.load(fp)
    if isinstance(new_info, str):
        with open(new_info, 'r') as fp:
            new_info = json.load(fp)

    for module_key in new_info.keys():
        if module_key in prev_info:
            prev_info[module_key].update(new_info[module_key])
        else:
            prev_info[module_key] = new_info[module_key]
    return prev_info


def save_profiled_results(models, save_path, detail, metrics = ["latency"]):
    new_models = merge_info(new_info=models, info_save_path=save_path)
    from .backend_meta.utils import dump_profiled_results
    with open(save_path, 'w') as fp:
        json.dump(dump_profiled_results(new_models, detail=detail, metrics=metrics), fp, indent=4)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import tensorflow as tf
from tensorflow.python.framework.convert_to_constants import convert_variables_to_constants_v2
import os


def keras_model_to_frozenpb(model, frozen_out_path, frozen_graph_filename, shapes, dtype=tf.dtypes.float32):
    full_model = tf.function(lambda x: model(x))
    if len(shapes) == 1:
        tensor_specs = tf.TensorSpec([1] + shapes[0], dtype)
    else:
        tensor_specs = [tf.TensorSpec([1] + shape, dtype) for shape in shapes]
    full_model = full_model.get_concrete_function(tensor_specs)

    frozen_func = convert_variables_to_constants_v2(full_model)
    frozen_func.graph.as_graph_def()

    frozen_graph_filename_bin = frozen_graph_filename + '.pb'
    frozen_graph_filename_txt = frozen_graph_filename + '.pbtxt'

    tf.io.write_graph(
        graph_or_graph_def=frozen_func.graph,
        logdir=frozen_out_path,
        name=frozen_graph_filename_bin,
        as_text=False
    )
    tf.io.write_graph(
        graph_or_graph_def=frozen_func.graph,
        logdir=frozen_out_path,
        name=frozen_graph_filename_txt,
        as_text=True
    )

    return (
        os.path.join(frozen_out_path, frozen_graph_filename_bin),
        os.path.join(frozen_out_path, frozen_graph_filename_txt),
    )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .openvino_backend import OpenVINOBackend
from .openvino_profiler import OpenVINOProfiler
from .vpu import OpenVINOVPUBackend



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .data_sampler import generate_config_sample, BaseConfigSampler
from .predictor_builder import build_predictor_by_data, BaseFeatureParser, collect_kernel_data


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .config_manager import (
    create_user_configs,
    get_user_data_folder,
    change_user_data_folder
)
from .utils import download_from_url


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import sys
import yaml
import logging
import importlib


__user_config_folder__ = os.path.expanduser('~/.nn_meter/config')
__registry_cfg_filename__ = 'registry.yaml'
__predictors_cfg_filename__ = 'predictors.yaml'

    
def import_module(package_location, module_path, class_name):
    sys.path.append(package_location)
    module = importlib.import_module(module_path)   
    backend_cls = getattr(module, class_name)
    return backend_cls


def register_module(module_type, meta_file):
    with open(meta_file, "r") as fp:
        meta_data = yaml.load(fp, yaml.FullLoader)

    # load module information to '~/.nn_meter/config/registry.yaml'
    builtin_name = meta_data.pop("builtin_name")
    import_module(meta_data["package_location"], meta_data["class_module"], meta_data["class_name"])

    # check necessary feature and run test script
    # for backend, check if there exits the default config file:
    if module_type == "backends":
        if meta_data["defaultConfigFile"] != None and not os.path.isfile(meta_data["defaultConfigFile"]):
            raise ValueError(f"The default config file {meta_data['defaultConfigFile']} does not exist")
    elif module_type == "kernels":
        import_module(meta_data["package_location"], meta_data["sampler_module"], meta_data["sampler_name"])
        import_module(meta_data["package_location"], meta_data["parser_module"], meta_data["parser_name"])
    if module_type == "operators" or module_type == "testcases":
        implement = meta_data.pop("implement")
        meta_data = {implement: meta_data}

    prev_info = {}
    if os.path.isfile(os.path.join(__user_config_folder__, __registry_cfg_filename__)):
        with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'r') as fp:
            prev_info = yaml.load(fp, yaml.FullLoader)

    if module_type in prev_info:
        if (module_type == "operators" or module_type == "testcases") and builtin_name in prev_info[module_type]:
            prev_info[module_type][builtin_name].update(meta_data)
        else:
            prev_info[module_type][builtin_name] = meta_data
    else:
        prev_info[module_type] = {builtin_name: meta_data}

    with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'w') as fp:
        yaml.dump(prev_info, fp)
    logging.keyinfo(f"Successfully register {module_type[:-1]}: {builtin_name}")


def register_predictor(meta_file):
    with open(meta_file, "r") as fp:
        meta_data = yaml.load(fp, yaml.FullLoader)
    
    # for predictors registration, load predictor information to '~/.nn_meter/config/predictors.yaml'
    builtin_name = meta_data["name"]
    with open(os.path.join(__user_config_folder__, __predictors_cfg_filename__), 'r') as fp:
        prev_info = yaml.load(fp, yaml.FullLoader)
    prev_info.append(meta_data)
    
    with open(os.path.join(__user_config_folder__, __predictors_cfg_filename__), 'w') as fp:
        yaml.dump(prev_info, fp)
    logging.keyinfo(f"Successfully register predictor: {builtin_name}")


def register_module_cli(args):
    if args.predictor:
        register_predictor(args.predictor)
    elif args.backend:
        register_module("backends", args.backend)
    elif args.operator:
        register_module("operators", args.operator)
    elif args.testcase:
        register_module("testcases", args.testcase)
    elif args.kernel:
        register_module("kernels", args.kernel)
    else:
        logging.keyinfo('please run "nn-meter register --help" to see guidance.')


def unregister_module(module_type, module_name):
    success = False
    if os.path.isfile(os.path.join(__user_config_folder__, __registry_cfg_filename__)):
        with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'r') as fp:
            prev_info = yaml.load(fp, yaml.FullLoader)
        if module_name in prev_info[module_type]:
            prev_info[module_type].pop(module_name)
            with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'w') as fp:
                yaml.dump(prev_info, fp)
            success = True
            logging.keyinfo(f"Successfully unregister {module_name}.")
    if not success:
        logging.keyinfo(f"Unregister failed: there is no module named {module_name}.")


def unregister_module_with_implement(module_type, module_name, module_implement):
    success = False
    if os.path.isfile(os.path.join(__user_config_folder__, __registry_cfg_filename__)):
        with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'r') as fp:
            prev_info = yaml.load(fp, yaml.FullLoader)
        try:
            prev_info[module_type][module_name].pop(module_implement)
            with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'w') as fp:
                yaml.dump(prev_info, fp)
            success = True
            logging.keyinfo(f"Successfully unregister {module_name} ({module_implement}).")
        except:
            pass
    if not success:
        logging.keyinfo(f"Unregister failed: there is no module named {module_name} ({module_implement}).")


def unregister_predictor(predictor_name, predictor_version):
    success = False
    with open(os.path.join(__user_config_folder__, __predictors_cfg_filename__), 'r') as fp:
        prev_info = yaml.load(fp, yaml.FullLoader)
    for i, p in enumerate(prev_info):
        if p['name'] == predictor_name and (predictor_version is None or p['version'] == predictor_version):
            prev_info.pop(i)
            success = True
    if success:
        with open(os.path.join(__user_config_folder__, __predictors_cfg_filename__), 'w') as fp:
            yaml.dump(prev_info, fp)
        logging.keyinfo(f"Successfully unregister {predictor_name}.")
    else:
        logging.keyinfo(f"Unregister failed: there is no predictor named {predictor_name}.")


def unregister_module_cli(args):
    if args.predictor:
        unregister_predictor(args.predictor, args.predictor_version)
    elif args.backend:
        unregister_module("backends", args.backend)
    elif args.operator:
        unregister_module_with_implement("operators", args.operator[0], args.operator[1])
    elif args.testcase:
        unregister_module_with_implement("testcases", args.testcase[0], args.testcase[1])
    elif args.kernel:
        unregister_module("kernels", args.kernel)
    else:
        logging.keyinfo('please run "nn-meter unregister --help" to see guidance.')



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import random


def sample_in_range(mind, maxd, sample_num):
    '''sample #sample_num data from a range [mind, maxd)
    '''
    # if the sample_num is bigger than sample population, we only keep the number of population to avoid repetition
    if maxd - mind <= sample_num:
        data = list(range(mind, maxd))
        random.shuffle(data)
        return data
    else:
        return random.sample(range(mind, maxd), sample_num)


def sample_cin_cout(cin, cout, sample_num): 
    '''fine-grained sample #sample_num data in the cin and cout dimensions, respectively
    '''
    cins = sample_in_range(int(cin * 0.5), int(cin * 1.2), sample_num)
    couts = sample_in_range(int(cout * 0.5), int(cout * 1.2), sample_num)
    l = min(len(cins), len(couts)) # align the length of cins and couts
    cins, couts = cins[:l], couts[:l]
    return cins, couts


def finegrained_sampling_conv(cfgs, count):
    ''' 
    Sampling configs for conv kernels
    Returned params include: (hw, cin, cout, kernel_size, strides)
    '''
    ncfgs = []
    for cfg in cfgs:
        cins, couts = sample_cin_cout(cfg['CIN'], cfg['COUT'], count)
        for cin, cout in zip(cins, couts):
            c = {
                'HW': cfg['HW'],
                'CIN': cin,
                'COUT': cout,
                'KERNEL_SIZE': cfg['KERNEL_SIZE'],
                'STRIDES': cfg['STRIDES'],
            }
            ncfgs.append(c)
    return ncfgs


def finegrained_sampling_dwconv(cfgs, count):
    ''' 
    Sampling configs for dwconv kernels
    Returned params include: (hw, cin, kernel_size, strides)
    '''
    ncfgs = []
    for cfg in cfgs:
        cins = sample_in_range(int(cfg['CIN'] * 0.5), int(cfg['CIN'] * 1.2), count)
        for cin in cins:
            c = {
                'HW': cfg['HW'],
                'CIN': cin,
                'KERNEL_SIZE': cfg['KERNEL_SIZE'],
                'STRIDES': cfg['STRIDES'],
            }
            ncfgs.append(c)
    return ncfgs


def finegrained_sampling_fc(cfgs, count):
    '''
    Sampling configs for fc kernels
    Returned params include: (cin, cout)
    '''
    ncfgs = []
    for cfg in cfgs:
        cins, couts = sample_cin_cout(cfg['CIN'], cfg['COUT'], count)
        for cin, cout in zip(cins, couts):
            c = {
                'CIN': cin,
                'COUT': cout
            }
            ncfgs.append(c)
    return ncfgs


def finegrained_sampling_pooling(cfgs, count):
    '''
    Sampling configs for pooling kernels
    Returned params include: (hw, cin, kernel_size, pool_strides)
    '''
    ncfgs = []
    for cfg in cfgs:
        cins = sample_in_range(int(cfg['CIN'] * 0.5), int(cfg['CIN'] * 1.2), count)
        for cin in cins:
            c = {
                'HW': cfg['HW'],
                'CIN': cin,
                'KERNEL_SIZE': cfg['KERNEL_SIZE'],
                'STRIDES': cfg['POOL_STRIDES'],
            }
            ncfgs.append(c)
    return ncfgs


def finegrained_sampling_hw_cin(cfgs, count):
    ''' sampling configs for kernels with hw and cin parameter
    Returned params include: (hw, cin)
    '''
    ncfgs = []
    for cfg in cfgs:
        cins = sample_in_range(int(cfg['CIN'] * 0.5), int(cfg['CIN'] * 1.2), count)
        for cin in cins:
            c = {
                'CIN': cin,
                'HW': cfg['HW']
            }
            ncfgs.append(c)
    return ncfgs


def finegrained_sampling_hw_cin_even(cfgs, count):
    ''' sampling configs for kernels with hw and cin (only even values) parameter, in case for split / se / channelshuffle
    Returned params include: (hw, cin)
    '''
    ncfgs = []
    for cfg in cfgs:
        cins = sample_in_range(int(cfg['CIN'] * 0.5), int(cfg['CIN'] * 1.2), count)
        for cin in cins:
            c = {
                'CIN': cin + 1 if cin % 2 else cin,
                'HW': cfg['HW']
            }
            ncfgs.append(c)
    return ncfgs


def finegrained_sampling_concats(cfgs, count):
    ''' sampling functions for concat kernel
    Returned params include: (hw, cin1, cin2, cin3, cin4). Note that we only sample (num of cin) = 2, 3, 4, 
    (cin1, cin2, cin3, cin4) is one-hot vector with unused input channel set as 0.
    '''
    ncfgs = []
    for cfg in cfgs:
        ncins, total_cins = [], []
        for cin in [cfg['CIN1'], cfg['CIN2'], cfg['CIN3'], cfg['CIN4']]:
            if cin == 0:
                total_cins.append([0] * count)
                continue
            cins = sample_in_range(int(cin * 0.5), int(cin * 1.2), count)
            ncins.append(len(cins))
            total_cins.append(cins)
        for j in range(min(ncins)):
            cins = [total_cins[i][j] for i in range(4)]
            c = {
                'HW': cfg['HW'],
                'CIN1': cins[0],
                'CIN2': cins[1],
                'CIN3': cins[2],
                'CIN4': cins[3]
            }
            ncfgs.append(c)
    return ncfgs



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import numpy as np
from sklearn.metrics import mean_squared_error


def get_kernel_name(optype):
    """
    for many similar kernels, we use one kernel predictor since their latency difference is negligible,
    return the kernel name via the optype

    """
    if "conv" in optype and "dwconv" not in optype:
        optype = "conv-bn-relu"
    if "dwconv" in optype:
        optype = "dwconv-bn-relu"
    if optype == "fc-relu":
        optype = "fc"
    if optype == "max-pool":
        optype = "maxpool"
    if optype == "avg-pool":
        optype = "avgpool"
    if optype in ["global-pool", "gap"]:
        optype = "global-avgpool"
    if optype == "channel_shuffle":
        optype = "channelshuffle"
    if optype in ["bn-relu"]:
        optype = "bnrelu"
    if optype in ["add-relu"]:
        optype = "addrelu"

    if optype in ["SE", "SE-relu", "se", "se-relu"]:
        optype = "se"

    return optype


def get_accuracy(y_pred, y_true, threshold=0.01):
    a = (y_true - y_pred) / y_true
    b = np.where(abs(a) <= threshold)
    return len(b[0]) / len(y_true)


def latency_metrics(y_pred, y_true):
    """
    evaluation metrics for prediction performance
    """
    y_true=np.array(y_true)
    y_pred=np.array(y_pred)
    rmspe = (np.sqrt(np.mean(np.square((y_true - y_pred) / y_true)))) * 100
    rmse = np.sqrt(mean_squared_error(y_pred, y_true))
    acc5 = get_accuracy(y_pred, y_true, threshold=0.05)
    acc10 = get_accuracy(y_pred, y_true, threshold=0.10)
    acc15 = get_accuracy(y_pred, y_true, threshold=0.15)
    return rmse, rmspe, rmse / np.mean(y_true), acc5, acc10, acc15


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import json
import logging
from .onnx_converter import OnnxConverter
from .frozenpb_converter import FrozenPbConverter
from .torch_converter import NNIBasedTorchConverter, OnnxBasedTorchConverter, NNIIRConverter
from nn_meter.utils.import_package import try_import_onnx, try_import_torch, try_import_torchvision_models
logging = logging.getLogger("nn-Meter")


def model_file_to_graph(filename: str, model_type: str, input_shape=(1, 3, 224, 224), apply_nni=False):
    """
    read the given file and convert the model in the file content to nn-Meter IR graph object 
    @params:
    filename: string to specify the location of the file
    
    model: the model to be predicted, allowed file format include
        - the path to a saved tensorflow model file (*.pb), `model_type` must be set to "pb"
        - string to specify the name of a built-in torch model from the torchvision model zoo, `model_type` must be set to "torch"
        - the path to a saved ONNX model file (*.onnx), `model_type` must be set to "onnx"
        - the path to a saved dictionary object following nn-Meter-IR format (*.json), `model_type` must be set to "nnmeter-ir"
        - the path to a saved dictionary object following NNI-IR format(*.json), `model_type` must be set to "nni-ir"
        
    model_type:  string to specify the type of parameter model, allowed items are ["pb", "torch", "onnx", "nnmeter-ir", "nni-ir"]
    
    input_shape: the shape of input tensor for inference (if necessary), a random tensor according to the shape will be generated and used. This parameter is only 
        accessed when model_type == 'torch'
    
    apply_nni: switch the torch converter used for torch model parsing. If apply_nni==True, NNI-based converter is used for torch model conversion, which requires 
        nni>=2.4 installation and should use nn interface from NNI `import nni.retiarii.nn.pytorch as nn` to define the PyTorch modules. Otherwise Onnx-based torch 
        converter is used, which requires onnx installation (well tested version is onnx==1.9.0). NNI-based converter is much faster while the conversion is unstable 
        as it could fail in some case. Onnx-based converter is much slower but stable compared to NNI-based converter. This parameter is only accessed when 
        model_type == 'torch'
    """
    if model_type == "onnx":
        onnx = try_import_onnx()
        model = onnx.load(filename)
        return onnx_model_to_graph(model)

    elif model_type == "pb":
        converter = FrozenPbConverter(filename)
        return converter.get_flatten_graph()

    elif model_type == "nni-ir":
        with open(filename, "r") as fp:
            model = json.load(fp)
        return nni_model_to_graph(model)

    elif model_type == "nnmeter-ir":
        with open(filename, "r") as fp:
            return json.load(fp)

    elif model_type == "torch":
        models = try_import_torchvision_models()
        torchvision_zoo_dict = {
            'resnet18': 'models.resnet18()',
            'alexnet': 'models.alexnet()',
            'vgg16': 'models.vgg16()',
            'squeezenet': 'models.squeezenet1_0()',
            'densenet161': 'models.densenet161()',
            'inception_v3': 'models.inception_v3()',
            'googlenet': 'models.googlenet()',
            'shufflenet_v2': 'models.shufflenet_v2_x1_0()',
            'mobilenet_v2': 'models.mobilenet_v2()',
            'resnext50_32x4d': 'models.resnext50_32x4d()',
            'wide_resnet50_2': 'models.wide_resnet50_2()',
            'mnasnet': 'models.mnasnet1_0()',
        }
        if filename in torchvision_zoo_dict:
            model = eval(torchvision_zoo_dict[filename])
        else:
            suppost_list = ", ".join([k for k in torchvision_zoo_dict])
            raise ValueError(f"Unsupported model name: {filename} in torchvision. Supporting list: {suppost_list}")
        return torch_model_to_graph(model, input_shape, apply_nni)

    else:
        raise ValueError(f"Unsupported model type: {model_type}")


def model_to_graph(model, model_type, input_shape=(1, 3, 224, 224), apply_nni=False):
    """
    convert the given model to nn-Meter IR graph object 
    @params:
    model: the model object for converting, allowed file format include
        - pytorch model object (nn.Module), `model_type` must be set to "torch"
        - ONNX model object, `model_type` must be set to "onnx"
        - dictionary object following NNI-IR format, `model_type` must be set to "nni-ir"
        
    model_type:  string to specify the type of parameter model, allowed items are ["torch", "onnx", "nnmeter-ir", "nni-ir"]
    
    input_shape: the shape of input tensor for inference (if necessary), a random tensor according to the shape will be generated and used. This parameter is only 
        accessed when model_type == 'torch'
    """
    if model_type == "onnx":
        return onnx_model_to_graph(model)
    elif model_type == "torch":
        return torch_model_to_graph(model, input_shape, apply_nni)
    elif model_type == "nni-ir":
        return nni_model_to_graph(model)
    elif model_type == "nnmeter-ir":
        return model # nnmeter-ir doesn't need any post-process
    else:
        raise ValueError(f"Unsupported model type: {model_type}")


def onnx_model_to_graph(model):
    converter = OnnxConverter(model)
    return converter.convert()


def nni_model_to_graph(model):
    converter = NNIIRConverter(model)
    return converter.convert()


def torch_model_to_graph(model, input_shape=(1, 3, 224, 224), apply_nni=False):
    torch = try_import_torch()
    args = torch.randn(*input_shape)
    try:
        # if the test model has no parameters (such as activation ops), there will be error when calling ``model.parameters``
        if next(model.parameters()).is_cuda:
            args = args.to("cuda")
    except:
        pass
    if apply_nni: 
        # apply NNI-based torch converter, which requires nni>=2.4 installation and should use nn interface from NNI 
        # `import nni.retiarii.nn.pytorch as nn` to define the PyTorch modules.
        try:
            logging.info("NNI-based Torch Converter is applied for model conversion")
            converter = NNIBasedTorchConverter(model, args)
        except:
            raise NotImplementedError("Your model is not fully converted by NNI-based converter. Please set apply_nni=False and try again.")
    else:
        # apply Onnx-based torch converter, which requires onnx installation (well tested version is onnx==1.9.0) 
        # and the conversion is more stable
        logging.info("Onnx-based Torch Converter is applied for model conversion")
        converter = OnnxBasedTorchConverter(model, args)
    return converter.convert()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import yaml
import pkg_resources
from shutil import copyfile

__backend_tflite_cfg_filename__ = 'backend_tflite_config.yaml'
__backend_openvino_cfg_filename__ = 'backend_openvino_config.yaml'
__ruletest_cfg_filename__ = 'ruletest_config.yaml'
__predbuild_cfg_filename__ = 'predictorbuild_config.yaml'


def copy_to_workspace(backend_type, workspace_path, backendConfigFile = None):
    """copy the default config file to user's workspace
    """
    os.makedirs(os.path.join(workspace_path, 'configs'), exist_ok=True)

    # backend config
    if backend_type == 'customized' and backendConfigFile:
        copyfile(backendConfigFile, os.path.join(workspace_path, 'configs', 'backend_config.yaml'))
    else:
        if backend_type == 'tflite':
            config_name = __backend_tflite_cfg_filename__
        elif backend_type == 'openvino':
            config_name = __backend_openvino_cfg_filename__
        copyfile(
            pkg_resources.resource_filename(".".join(__name__.split('.')[:-2]), 'configs/builder/backends/' + config_name), 
            os.path.join(workspace_path, 'configs', 'backend_config.yaml'))
    # rule test config
    copyfile(
        pkg_resources.resource_filename(".".join(__name__.split('.')[:-2]), f'configs/builder/fusion_rule_tester/' + __ruletest_cfg_filename__), 
        os.path.join(os.path.join(workspace_path, 'configs'), 'ruletest_config.yaml'))
    # predictor builder config
    copyfile(
        pkg_resources.resource_filename(".".join(__name__.split('.')[:-2]), f'configs/builder/predictor_builder/' + __predbuild_cfg_filename__), 
        os.path.join(os.path.join(workspace_path, 'configs'), 'predictorbuild_config.yaml'))


def load_config_file(workspace_path):
    """load config file from workspace_path;
    if the file not located in workspace_path, copy it from distribution
    """
    backend_filepath = os.path.join(workspace_path, "configs", 'backend_config.yaml')
    ruletest_filepath = os.path.join(workspace_path, "configs", 'ruletest_config.yaml')
    predictorbuild_filepath = os.path.join(workspace_path, "configs", 'predictorbuild_config.yaml')
    try:
        try:
            with open(backend_filepath) as fp:
                backend = yaml.load(fp, yaml.FullLoader)
        except: # In customized backend, sometimes there is no backend config file, skip loading config for this case
            backend = None
        with open(ruletest_filepath) as fp:
            ruletest = yaml.load(fp, yaml.FullLoader)
        with open(predictorbuild_filepath) as fp:
            predictorbuild = yaml.load(fp, yaml.FullLoader)
        return backend, ruletest, predictorbuild
    except:
        raise FileNotFoundError(f"config file in {workspace_path} not found, please create the workspace first." \
                                "Docs: https://github.com/microsoft/nn-Meter/blob/main/docs/builder/overview.md#-create-workspace-")


class ConfigData:
    def __init__(self):
        self.workspace_path = ''
        self._global_settings = {}

    def set(self, name, value, module=''):
        self._global_settings[module][name] = value
    
    def set_module(self, value, module=''):
        self._global_settings[module] = value
        
    def get(self, name, module=''):
        try:
            return self._global_settings[module].get(name)
        except:
            raise ValueError(f"Could not find {module} or {module}.{name} in builder config. \
                Please run `builder_config.init('path/to/workspace')` first.")

    def get_module(self, module=''):
        try:
            return self._global_settings[module]
        except:
            raise ValueError(f"Module [{module}] from config file not found, please create the workspace first. " \
                "Docs: https://github.com/microsoft/nn-Meter/blob/main/docs/builder/overview.md#-create-workspace-")

    def get_settings(self):
        return self._global_settings


class ConfigManager(ConfigData):
    def init(self, workspace_path):
        self.workspace_path = workspace_path
        self._load_from_config_file(workspace_path)
    
    def _load_from_config_file(self, workspace_path):
        backend, ruletest, predbuild = load_config_file(workspace_path)
        self.set_module(backend, 'backend')
        self.set_module(ruletest, 'ruletest')
        self.set_module(predbuild, 'predbuild')
        self.set('WORKSPACE', os.path.join(self.workspace_path, "fusion_rule_test"), 'ruletest')
        self.set('WORKSPACE', os.path.join(self.workspace_path, "predictor_build"), 'predbuild')

builder_config = ConfigManager()
